/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007-2014 Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix@posteo.de

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"

/**************************************************************************
   Fotoxx image edit - Image menu functions
***************************************************************************/


/**************************************************************************/

//  Rotate image right or left 90 degrees to upright the image.
//  This is not an edit transaction: file is replaced and re-opened.

zdialog  *zdrotate90;

void m_rotate90(GtkWidget *, cchar *menu)                                  //  v.13.08
{
   int  turnevent(zdialog *, cchar *);
   void turnfunc(GtkWidget *, cchar *menu);

   if (! curr_file) return;                                                //  v.13.09
   F1_help_topic = "rotate";
   
   if (checkpend("all")) return;                                           //  v.13.12
   Fmenulock = 1;

   zdrotate90 = zdialog_new(ZTX("Rotate Image 90º"),mWin,null);
   zdialog_add_widget(zdrotate90,"vbox","vb1","dialog",0,"space=3");
   GtkWidget *tbar = create_toolbar(zdialog_widget(zdrotate90,"vb1"),32);
   gtk_toolbar_set_style(GTK_TOOLBAR(tbar),GTK_TOOLBAR_BOTH);
   add_toolbar_button(tbar,"left",ZTX("rotate -90º"),"rotate-left.png",turnfunc);
   add_toolbar_button(tbar,"right",ZTX("rotate +90º"),"rotate-right.png",turnfunc);
   add_toolbar_button(tbar,"cancel",0,"cancel.png",turnfunc);
   zdialog_resize(zdrotate90,200,0);
   zdialog_run(zdrotate90,turnevent,"save");
   zdialog_wait(zdrotate90);

   Fmenulock = 0;
   return;
}


int turnevent(zdialog *zd, cchar *event)
{
   if (zd->zstat) zdialog_free(zd);                                        //  [x] button
   return 1;
}


void turnfunc(GtkWidget *, cchar *menu)
{
   int      angle = 0;

   if (strEqu(menu,"cancel")) {
      zdialog_free(zdrotate90);
      return;
   }

   if (strEqu(menu,"right")) angle = +90;
   if (strEqu(menu,"left")) angle = -90;

   mutex_lock(&Fpixmap_lock);                                              //  block window updates
   E0pxm = PXM_load(curr_file,1);                                          //  load poss. 16-bit image         v.12.10
   if (E0pxm) {
      E3pxm = PXM_rotate(E0pxm,angle);                                     //  rotate
      PXM_free(E0pxm);
      E0pxm = E3pxm;
      E3pxm = 0;
   }
   mutex_unlock(&Fpixmap_lock);

   if (strEqu(curr_file_type,"other"))                                     //  if unsupported type, use jpg
      strcpy(curr_file_type,"jpg");

   f_save(curr_file,curr_file_type,curr_file_bpc);                         //  save file with curr. edits applied
   f_open_saved();                                                         //  update tab F                    v.13.10

   return;
}


/**************************************************************************/

//  Rotate image right or left 90 degrees to upright the image.
//  This version is for clicked gallery thumbnails only.
//  This is not an edit transaction: file is replaced on disk.

void m_rotate90_thumb(GtkWidget *, cchar *)                                //  v.13.10
{
   int  turnevent_thumb(zdialog *zd, cchar *event);
   void  turnfunc_thumb(GtkWidget *, cchar *menu);

   int      err, Nth = 0;
   char     *turn_file;

   if (! clicked_file) return;
   turn_file = clicked_file;
   clicked_file = 0;
   
   if (checkpend("all")) return;                                           //  v.13.12

   err = f_open(turn_file,Nth,0,1);                                        //  open clicked image file
   free(turn_file);
   if (err) return;

   Fmenulock = 1;                                                          //  block other edit funcs       v.13.12

   zdrotate90 = zdialog_new(ZTX("Rotate Image 90º"),mWin,null);
   zdialog_add_widget(zdrotate90,"vbox","vb1","dialog",0,"space=3");
   GtkWidget *tbar = create_toolbar(zdialog_widget(zdrotate90,"vb1"),32);
   gtk_toolbar_set_style(GTK_TOOLBAR(tbar),GTK_TOOLBAR_BOTH);
   add_toolbar_button(tbar,"left",ZTX("rotate -90º"),"rotate-left.png",turnfunc_thumb);
   add_toolbar_button(tbar,"right",ZTX("rotate +90º"),"rotate-right.png",turnfunc_thumb);
   add_toolbar_button(tbar,"cancel",0,"gtk-cancel",turnfunc_thumb);
   zdialog_resize(zdrotate90,200,0);
   zdialog_run(zdrotate90,turnevent_thumb,"save");
   zdialog_wait(zdrotate90);

   Fmenulock = 0;                                                          //  v.13.12
   gallery(curr_file,"paint");
   settab("G");
   return;
}


int turnevent_thumb(zdialog *zd, cchar *event)
{
   if (zd->zstat) zdialog_free(zd);                                        //  [x] button
   return 1;
}


void turnfunc_thumb(GtkWidget *, cchar *menu)
{
   int      angle = 0;

   if (strEqu(menu,"cancel")) {
      zdialog_free(zdrotate90);
      return;
   }

   if (strEqu(menu,"right")) angle = +90;
   if (strEqu(menu,"left")) angle = -90;
   zdialog_free(zdrotate90);

   E0pxm = PXM_load(curr_file,1);                                          //  load poss. 16-bit image
   if (E0pxm) {
      E3pxm = PXM_rotate(E0pxm,angle);                                     //  rotate
      PXM_free(E0pxm);
      E0pxm = E3pxm;
      E3pxm = 0;
   }

   if (strEqu(curr_file_type,"other"))                                     //  if unsupported type, use jpg
      strcpy(curr_file_type,"jpg");

   f_save(curr_file,curr_file_type,curr_file_bpc);                         //  save rotated file
   f_open_saved();                                                         //  update tab F                 v.13.10

   strcpy(curr_file_type,f_save_type);                                     //  update curr_file_xxx from f_save_xxx
   curr_file_size = f_save_size;
   PXM_free(E0pxm);

   return;
}


/**************************************************************************/

//  trim (crop) and/or rotate an image                                     //  combine trim and rotate      v.13.10

namespace trimrotate_names
{
// int      trimrect[4];                                                   //  current trim rectangle (GLOBAL fotoxx.h)
// int      trimsize[2];                                                   //  current trim size (GLOBAL fotoxx.h)
   int      ptrimrect[4];                                                  //  prior trim rectangle (this image)
   int      ptrimsize[2];                                                  //  prior trim size (previous image)
   float    trimR;                                                         //  trim ratio, width/height
   float    rotate_goal, rotate_angle;                                     //  target and actual rotation
   int      rotatecrop1 = 0, rotatecrop2;                                  //  default rotate with no autocrop
   int      E0ww, E0hh, E3ww, E3hh;                                        //  full size image and preview size
   float    FPR;                                                           //  full size / preview ratio
   int      Fguidelines, guidelineX, guidelineY;                           //  horz/vert guidelines for rotate

   editfunc EFtrimrotate;

   void   trimrotate_initz(int autotrim);
   void   trimrotate_dialog();
   int    trimrotate_dialog_event(zdialog *zd, cchar *event);
   void   trim_customize();
   void   trimrotate_mousefunc();
   void   trim_darkmargins();
   void   trim_final();
   void   rotate_func();
   void   trimrotate_final();
   void   trimrotate_drawlines();
}


//  menu function

void m_trimrotate(GtkWidget *, cchar *menu)
{
   using namespace trimrotate_names;

   int         ii, xmargin, ymargin;
   cchar       *trim_message = ZTX("Trim: drag middle to move, drag corners to resize");
   cchar       *rotate_mess = ZTX("Minor rotate: drag right edge with mouse");
   char        text[20];
   zdialog     *zd;

   F1_help_topic = "trim_rotate";

   EFtrimrotate.menufunc = m_trimrotate;                                   //  menu function
   EFtrimrotate.funcname = "trim_rotate";
   EFtrimrotate.FprevReq = 1;                                              //  use preview
   EFtrimrotate.Frestart = 1;                                              //  allow restart
   EFtrimrotate.mousefunc = trimrotate_mousefunc;
   if (! edit_setup(EFtrimrotate)) return;                                 //  setup edit

   E0ww = E0pxm->ww;                                                       //  full-size image dimensions
   E0hh = E0pxm->hh;
   E3ww = E3pxm->ww;                                                       //  preview image (possibly smaller)
   E3hh = E3pxm->hh;
   FPR = 1.0 * (E0ww + E0hh) / (E3ww + E3hh);                              //  full image / preview ratio

   if (menu && strEqu(menu,"autotrim")) {                                  //  autotrim: keep preset trim rectangle
      trimsize[0] = trimrect[2] - trimrect[0];                             //  (full image scale)
      trimsize[1] = trimrect[3] - trimrect[1];
      trimR = 1.0 * trimsize[0] / trimsize[1];                             //  trim ratio = width/height
   }

   else
   {
      trimsize[0] = 0.8 * E0ww;                                            //  initial trim rectangle, 80% image size
      trimsize[1] = 0.8 * E0hh;
      xmargin = 0.5 * (E0ww - trimsize[0]);                                //  set balanced margins
      ymargin = 0.5 * (E0hh - trimsize[1]);
      trimrect[0] = xmargin;
      trimrect[2] = E0ww - xmargin;
      trimrect[1] = ymargin;
      trimrect[3] = E0hh - ymargin;
      trimR = 1.0 * trimsize[0] / trimsize[1];                             //  trim ratio = width/height
   }

   trimrect[0] = trimrect[0] / FPR;                                        //  use preview scale from now on
   trimrect[1] = trimrect[1] / FPR;
   trimrect[2] = trimrect[2] / FPR;
   trimrect[3] = trimrect[3] / FPR;

   ptrimrect[0] = 0;                                                       //  set prior trim rectangle
   ptrimrect[1] = 0;                                                       //    = 100% of image
   ptrimrect[2] = E3ww;
   ptrimrect[3] = E3hh;

   rotate_goal = rotate_angle = 0;                                         //  initially no rotation
   rotatecrop2 = rotatecrop1;                                              //  rotate autocrop setting

/***
       _______________________________________________________
      |                                                       |
      |  Trim: drag middle to move, drag corners to resize    |
      |                                                       |
      |  width [____|-|+]  height [____|-|+]   ratio 1.5      |
      |  trim size: [Max] [Prev] [Invert]  [x] Lock Ratio     |
      |  [1:1] [2:1] [3:2] [4:3] [16:9] [gold] [Customize]    |
      |                                                       |
      |  Minor rotate: drag right edge with mouse             |
      |  Rotate: degrees [____|-+]    [x] auto-trim           |
      |  [left 90]  [right 90]                                |
      |                                                       |
      |                       [Grid] [Apply] [Done] [Cancel]  |
      |_______________________________________________________|

***/

   if (! EFtrimrotate.zd)
   {
      zd = zdialog_new(ZTX("Trim/Rotate"),mWin,Bgrid,Bapply,Bdone,Bcancel,null);
      EFtrimrotate.zd = zd;

      zdialog_add_widget(zd,"label","labtrim","dialog",trim_message,"space=3");
      zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","labW","hb1",ZTX("width"),"space=5");
      zdialog_add_widget(zd,"spin","width","hb1","20|9999|1|1000");
      zdialog_add_widget(zd,"label","space","hb1",0,"space=5");
      zdialog_add_widget(zd,"label","labH","hb1",ZTX("height"),"space=5");
      zdialog_add_widget(zd,"spin","height","hb1","20|9999|1|600");
      zdialog_add_widget(zd,"label","space","hb1",0,"space=5");
      zdialog_add_widget(zd,"label","labR","hb1",ZTX("ratio"),"space=5");
      zdialog_add_widget(zd,"label","ratio","hb1","1.67   ");

      zdialog_add_widget(zd,"hbox","hb2","dialog");
      zdialog_add_widget(zd,"label","lab2","hb2",ZTX("trim size:"),"space=5");
      zdialog_add_widget(zd,"button","max","hb2",Bmax,"space=5");
      zdialog_add_widget(zd,"button","prev","hb2",Bprev,"space=5");
      zdialog_add_widget(zd,"button","invert","hb2",Binvert,"space=5");
      zdialog_add_widget(zd,"check","lock","hb2",ZTX("Lock Ratio"),"space=15");

      zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=4");                           //  ratio buttons inserted here
      for (ii = 0; ii < 6; ii++)
         zdialog_add_widget(zd,"button",trimbuttons[ii],"hb3",trimbuttons[ii],"space=5");
      zdialog_add_widget(zd,"button","custom","hb3",ZTX("Customize"),"space=5");          //  and [custom] button

      zdialog_add_widget(zd,"hsep","sep1","dialog");

      zdialog_add_widget(zd,"label","labrotmess","dialog",rotate_mess,"space=5");
      zdialog_add_widget(zd,"hbox","hb5","dialog");
      zdialog_add_widget(zd,"label","labrotate","hb5",ZTX("Rotate: degrees"),"space=5");
      zdialog_add_widget(zd,"spin","degrees","hb5","-360|360|0.1|0");
      zdialog_add_widget(zd,"check","rotatecrop","hb5",ZTX("auto-trim"),"space=20");
      zdialog_add_widget(zd,"hbox","hb6","dialog",0,"space=2");
      zdialog_add_widget(zd,"label","lableft","hb6",ZTX("left"),"space=5");
      zdialog_add_widget(zd,"button","-90","hb6","-90");
      zdialog_add_widget(zd,"label","space","hb6",0,"space=10");
      zdialog_add_widget(zd,"label","labr","hb6",ZTX("right"),"space=5");
      zdialog_add_widget(zd,"button","+90","hb6","+90");
   }

   zd = EFtrimrotate.zd;

   zdialog_stuff(zd,"width",trimsize[0]);                                  //  stuff width, height, ratio
   zdialog_stuff(zd,"height",trimsize[1]);
   sprintf(text,"%.2f  ",trimR);
   zdialog_stuff(zd,"ratio",text);
   zdialog_stuff(zd,"degrees",0);
   zdialog_stuff(zd,"rotatecrop",rotatecrop1);                             //  preserve rotate autocrop setting

   zdialog_run(zd,trimrotate_dialog_event,"save");                         //  run dialog - parallel
   takeMouse(trimrotate_mousefunc,dragcursor);                             //  connect mouse function
   
   currgrid = 1;                                                           //  use trim/rotate grid            v.13.11
   
   PXM_free(E9pxm);
   E9pxm = PXM_copy(E3pxm);
   Fzoom = 0;
   trim_darkmargins();

   return;
}


//  dialog event and completion callback function

int trimrotate_names::trimrotate_dialog_event(zdialog *zd, cchar *event)
{
   using namespace trimrotate_names;

   static int  flip = 0;
   int         width, height, delta;
   int         ii, rlock;
   float       r1, r2, ratio = 0;
   char        text[20];
   cchar       *pp;
   
   if (strEqu(event,"done")) zd->zstat = 3;                                //  apply and quit               v.13.12
   if (strEqu(event,"enter")) zd->zstat = 3;                               //  v.14.03

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) {                                                //  grid
         zd->zstat = 0;                                                    //  keep dialog active
         m_gridlines(0,"grid 1");
         return 1;
      }

      if (zd->zstat == 2) {                                                //  apply
         trimrotate_final();                                               //  use full-size images
         edit_done(0);
         m_trimrotate(0,0);                                                //  restart edit
         return 1;
      }

      if (zd->zstat == 3) {                                                //  done
         trimrotate_final();                                               //  use full-size images
         edit_done(0);
         currgrid = 0;                                                     //  restore normal grid settings
         return 1;
      }

      draw_toplines(2);                                                    //  erase trim rectangle
      erase_topcircles();                                                  //    and target circle
      currgrid = 0;                                                        //  restore normal grid settings
      edit_cancel(0);                                                      //  cancel or kill
      return 1;
   }

   if (strEqu(event,"custom")) {
      edit_cancel(0);                                                      //  cancel edit
      trim_customize();                                                    //  customize dialog
      m_trimrotate(0,0);                                                   //  restart edit
      return 1;
   }

   if (strEqu(event,"focus")) {
      takeMouse(trimrotate_mousefunc,dragcursor);                          //  connect mouse function
      return 1;
   }

   if (strEqu(event,"prev"))                                               //  use width/height from prior        v.14.01
   {                                                                       //    image trim 
      width = ptrimsize[0];
      height = ptrimsize[1];
      if (width < 20 || height < 20) return 1;                             //  no value established, ignore
      if (width > E0ww) width = E0ww;
      if (height > E0hh) height = E0hh;
      zdialog_stuff(zd,"width",width);
      zdialog_stuff(zd,"height",height);
      event = "width";                                                     //  process same as manual inputs
   }

   if (strstr("width height",event))                                       //  get direct width/height inputs
   {
      zdialog_fetch(zd,"width",width);                                     //  full image scale
      zdialog_fetch(zd,"height",height);

      if (width < trimsize[0])                                             //  increase full scale change as needed
         while (int(width/FPR) == int(trimsize[0]/FPR)) width--;           //    to make a change at preview scale

      if (width > trimsize[0])
         while (int(width/FPR) == int(trimsize[0]/FPR)) width++;

      if (height < trimsize[1])
         while (int(height/FPR) == int(trimsize[1]/FPR)) height--;

      if (height > trimsize[1])
         while (int(height/FPR) == int(trimsize[1]/FPR)) height++;

      zdialog_fetch(zd,"lock",rlock);                                      //  lock ratio on/off

      width = width / FPR;                                                 //  preview scale
      height = height / FPR;

      if (strEqu(event,"width")) {
         if (width > E3ww) width = E3ww;
         if (rlock) {                                                      //  ratio locked
            height = width / trimR + 0.5;                                  //  try to keep ratio
            if (height > E3hh) height = E3hh;
         }
      }

      if (strEqu(event,"height")) {
         if (height > E3hh) height = E3hh;
         if (rlock) {
            width = height * trimR + 0.5;
            if (width > E3ww) width = E3ww;
         }
      }

      flip = 1 - flip;                                                     //  alternates 0, 1, 0, 1 ...

      delta = width - int(trimsize[0]/FPR);

      if (delta > 0) {                                                     //  increased width
         trimrect[0] = trimrect[0] - delta / 2;                            //  left and right sides equally
         trimrect[2] = trimrect[2] + delta / 2;
         if (delta % 2) {                                                  //  if increase is odd
            trimrect[0] = trimrect[0] - flip;                              //    add 1 alternatively to each side
            trimrect[2] = trimrect[2] + 1 - flip;
         }
         if (trimrect[0] < 0) {                                            //  add balance to upper limit
            trimrect[2] = trimrect[2] - trimrect[0];
            trimrect[0] = 0;
         }
         if (trimrect[2] > E3ww) {                                         //  add balance to lower limit
            trimrect[0] = trimrect[0] - trimrect[2] + E3ww;
            trimrect[2] = E3ww;
         }
      }

      if (delta < 0) {                                                     //  decreased width
         trimrect[0] = trimrect[0] - delta / 2;
         trimrect[2] = trimrect[2] + delta / 2;
         if (delta % 2) {
            trimrect[0] = trimrect[0] + flip;
            trimrect[2] = trimrect[2] - 1 + flip;
         }
      }

      delta = height - int(trimsize[1]/FPR);

      if (delta > 0) {                                                     //  increased height
         trimrect[1] = trimrect[1] - delta / 2;                            //  top and bottom sides equally
         trimrect[3] = trimrect[3] + delta / 2;
         if (delta % 2) {                                                  //  if increase is odd
            trimrect[1] = trimrect[1] - flip;                              //    add 1 alternatively to each side
            trimrect[3] = trimrect[3] + 1 - flip;
         }
         if (trimrect[1] < 0) {
            trimrect[3] = trimrect[3] - trimrect[1];
            trimrect[1] = 0;
         }
         if (trimrect[3] > E3hh) {
            trimrect[1] = trimrect[1] - trimrect[3] + E3hh;
            trimrect[3] = E3hh;
         }
      }

      if (delta < 0) {                                                     //  decreased height
         trimrect[1] = trimrect[1] - delta / 2;
         trimrect[3] = trimrect[3] + delta / 2;
         if (delta % 2) {
            trimrect[1] = trimrect[1] + flip;
            trimrect[3] = trimrect[3] - 1 + flip;
         }
      }

      if (trimrect[0] < 0) trimrect[0] = 0;                                //  keep within limits
      if (trimrect[2] > E3ww) trimrect[2] = E3ww;                          //  use ww not ww-1
      if (trimrect[1] < 0) trimrect[1] = 0;
      if (trimrect[3] > E3hh) trimrect[3] = E3hh;

      width = trimrect[2] - trimrect[0];                                   //  new width and height
      height = trimrect[3] - trimrect[1];

      if (width > E3ww) width = E3ww;                                      //  limit to actual size      v.13.11
      if (height > E3hh) height = E3hh;

      width = width * FPR;                                                 //  full image scale
      height = height * FPR;

      trimsize[0] = width;                                                 //  new trim size
      trimsize[1] = height;

      zdialog_stuff(zd,"width",width);                                     //  update dialog values
      zdialog_stuff(zd,"height",height);

      if (! rlock)                                                         //  set new ratio if not locked
         trimR = 1.0 * trimsize[0] / trimsize[1];

      sprintf(text,"%.2f  ",trimR);                                        //  stuff new ratio
      zdialog_stuff(zd,"ratio",text);

      trim_darkmargins();                                                  //  show trim area in image
      return 1;
   }

   if (strEqu(event,"max"))                                                //  maximize trim rectangle            v.14.01
   {
      trimrect[0] = 0;
      trimrect[1] = 0;
      trimrect[2] = E3ww;
      trimrect[3] = E3hh;

      width = E3ww * FPR;                                                  //  full scale trim size
      height = E3hh * FPR;

      trimsize[0] = width;                                                 //  new trim size
      trimsize[1] = height;

      zdialog_stuff(zd,"width",width);                                     //  update dialog values
      zdialog_stuff(zd,"height",height);

      trimR = 1.0 * trimsize[0] / trimsize[1];

      sprintf(text,"%.2f  ",trimR);                                        //  stuff new ratio
      zdialog_stuff(zd,"ratio",text);
      zdialog_stuff(zd,"lock",0);                                          //  set no ratio lock

      trim_darkmargins();                                                  //  show trim area in image
      return 1;
   }

   if (strEqu(event,"invert"))                                             //  invert width/height dimensions     v.14.01
      ratio = 1.0 / trimR;                                                 //  mark ratio changed

   for (ii = 0; ii < 6; ii++)                                              //  trim ratio buttons
      if (strEqu(event,trimbuttons[ii])) break;
   if (ii < 6) {
      r1 = r2 = ratio = 0;
      pp = strField(trimratios[ii],':',1);
      if (pp) r1 = atof(pp);
      pp = strField(trimratios[ii],':',2);
      if (pp) r2 = atof(pp);
      if (r1 > 0 && r2 > 0) ratio = r1/r2;
      if (ratio < 0.1 || ratio > 10) ratio = 1.0;
      if (! ratio) return 1;
      zdialog_stuff(zd,"lock",1);                                          //  assume lock is wanted
      trimR = ratio;
   }
   
   if (ratio)                                                              //  ratio was changed
   {
      trimR = ratio;

      if (trimrect[2] - trimrect[0] > trimrect[3] - trimrect[1])
         trimrect[3] = trimrect[1] + (trimrect[2] - trimrect[0]) / trimR;  //  adjust smaller dimension
      else
         trimrect[2] = trimrect[0] + (trimrect[3] - trimrect[1]) * trimR;

      if (trimrect[2] > E3ww) {                                            //  if off the right edge,
         trimrect[2] = E3ww;                                               //  adjust height
         trimrect[3] = trimrect[1] + (trimrect[2] - trimrect[0]) / trimR;
      }

      if (trimrect[3] > E3hh) {                                            //  if off the bottom edge,
         trimrect[3] = E3hh;                                               //  adjust width
         trimrect[2] = trimrect[0] + (trimrect[3] - trimrect[1]) * trimR;
      }

      trimsize[0] = FPR * (trimrect[2] - trimrect[0]);                     //  new rectangle dimensions
      trimsize[1] = FPR * (trimrect[3] - trimrect[1]);

      zdialog_stuff(zd,"width",trimsize[0]);                               //  stuff width, height, ratio
      zdialog_stuff(zd,"height",trimsize[1]);
      sprintf(text,"%.2f  ",trimR);
      zdialog_stuff(zd,"ratio",text);

      trim_darkmargins();                                                  //  update trim area in image
      return 1;
   }

   if (strstr("+90 -90 degrees",event))                                    //  rotate action
   {
      if (strEqu(event,"+90"))                                             //  90 deg. CW
         rotate_goal += 90;

      if (strEqu(event,"-90"))                                             //  90 deg. CCW
         rotate_goal -= 90;

      if (strEqu(event,"degrees"))                                         //  degrees adjustment
         zdialog_fetch(zd,"degrees",rotate_goal);

      if (rotate_goal > 180) rotate_goal -= 360;                           //  keep within -180 to +180 deg.
      if (rotate_goal < -180) rotate_goal += 360;
      if (fabsf(rotate_goal) < 0.01) rotate_goal = 0;                      //  = 0 within my precision
      zdialog_stuff(zd,"degrees",rotate_goal);

      rotate_func();                                                       //  E3 is rotated E1
   }

   if (strEqu(event,"rotatecrop")) {
      zdialog_fetch(zd,"rotatecrop",rotatecrop1);                          //  new autocrop setting
      rotate_func();                                                       //  E3 is rotated E1
   }

   return 1;
}


//  dialog to get custom trim button names and corresponding ratios

void trimrotate_names::trim_customize()
{
   using namespace trimrotate_names;

   char        text[20], blab[8], rlab[8];
   float       r1, r2, ratio;
   cchar       *pp;
   int         ii, zstat;

   zdialog *zd = zdialog_new(ZTX("Trim Buttons"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hbb","dialog",0,"homog|space=3");
   zdialog_add_widget(zd,"hbox","hbr","dialog",0,"homog|space=3");
   zdialog_add_widget(zd,"label","label","hbb",ZTX("label"),"space=3");
   zdialog_add_widget(zd,"label","ratio","hbr",ZTX("ratio"),"space=3");

   strcpy(blab,"butt-0");
   strcpy(rlab,"ratio-0");

   for (ii = 0; ii < 6; ii++)                                              //  add current trimbuttons to dialog
   {
      blab[5] = '0' + ii;
      rlab[6] = '0' + ii;
      zdialog_add_widget(zd,"entry",blab,"hbb",trimbuttons[ii],"scc=6");
      zdialog_add_widget(zd,"entry",rlab,"hbr",trimratios[ii],"scc=6");
   }

   zdialog_run(zd,0,"mouse");                                              //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for complete

   if (zstat == 1)                                                         //  done
   {
      for (ii = 0; ii < 6; ii++)                                           //  get new button names
      {
         blab[5] = '0' + ii;
         zdialog_fetch(zd,blab,text,12);
         strTrim2(text);
         if (! *text) continue;
         free(trimbuttons[ii]);
         trimbuttons[ii] = strdup(text);
      }

      for (ii = 0; ii < 6; ii++)                                           //  get new ratios
      {
         rlab[6] = '0' + ii;
         zdialog_fetch(zd,rlab,text,12);
         strTrim2(text);
         r1 = r2 = ratio = 0;
         pp = strField(text,':',1);
         if (! pp) continue;
         r1 = atof(pp);
         pp = strField(text,':',2);
         if (! pp) continue;
         r2 = atof(pp);
         if (r1 > 0 && r2 > 0) ratio = r1/r2;
         if (ratio < 0.1 || ratio > 10) continue;
         free(trimratios[ii]);
         trimratios[ii] = strdup(text);
      }
   }

   zdialog_free(zd);                                                       //  kill dialog
   save_params();                                                          //  save parameters

   return;
}


//  trim/rotate mouse function

void trimrotate_names::trimrotate_mousefunc()
{
   using namespace trimrotate_names;

   static int  Fcorner, Frotate;
   int         mpx, mpy, xdrag, ydrag, rlock;
   int         chop, moveall = 0;
   int         dx, dy, dd, d1, d2, d3, d4;
   char        text[20];
   float       drr;
   zdialog     *zd = EFtrimrotate.zd;

   if (! LMclick && ! RMclick && ! Mxdrag && ! Mydrag) {                   //  no click or drag
      Fcorner = Frotate = 0;
      return;
   }

   if (LMclick)                                                            //  add vertical and horizontal
   {                                                                       //    lines at click position       v.13.10
      LMclick = 0;
      Fcorner = Frotate = 0;
      Fguidelines = 1;
      guidelineX = Mxclick;
      guidelineY = Myclick;
      trim_darkmargins();
      mwpaint2();
      return;
   }

   if (RMclick)                                                            //  remove the lines
   {
      RMclick = 0;
      Fcorner = Frotate = 0;
      Fguidelines = 0;
      trim_darkmargins();
      mwpaint2();
      return;
   }

   if (Mxdrag || Mydrag)                                                   //  drag
   {
      mpx = Mxdrag;
      mpy = Mydrag;
      xdrag = Mxdrag - Mxdown;
      ydrag = Mydrag - Mydown;
      Mxdown = Mxdrag;                                                     //  reset drag origin
      Mydown = Mydrag;
      Mxdrag = Mydrag = 0;
   }

   if (Frotate) goto mouse_rotate;                                         //  continue rotate in progress

   else if (Fcorner)
   {
      if (Fcorner == 1) { trimrect[0] = mpx; trimrect[1] = mpy; }          //  continue dragging corner with mouse
      if (Fcorner == 2) { trimrect[2] = mpx; trimrect[1] = mpy; }
      if (Fcorner == 3) { trimrect[2] = mpx; trimrect[3] = mpy; }
      if (Fcorner == 4) { trimrect[0] = mpx; trimrect[3] = mpy; }
   }

   else
   {
      moveall = 1;
      dd = 0.2 * (trimrect[2] - trimrect[0]);                              //  test if mouse is in the broad
      if (mpx < trimrect[0] + dd) moveall = 0;                             //    middle of the rectangle
      if (mpx > trimrect[2] - dd) moveall = 0;
      dd = 0.2 * (trimrect[3] - trimrect[1]);
      if (mpy < trimrect[1] + dd) moveall = 0;
      if (mpy > trimrect[3] - dd) moveall = 0;
   }

   if (moveall) {                                                          //  yes, move the whole rectangle
      trimrect[0] += xdrag;
      trimrect[2] += xdrag;
      trimrect[1] += ydrag;
      trimrect[3] += ydrag;
   }

   else if (! Fcorner)                                                     //  no, find closest corner to mouse
   {
      dx = mpx - trimrect[0];
      dy = mpy - trimrect[1];
      d1 = sqrt(dx*dx + dy*dy);                                            //  distance from NW corner

      dx = mpx - trimrect[2];
      dy = mpy - trimrect[1];
      d2 = sqrt(dx*dx + dy*dy);                                            //  NE

      dx = mpx - trimrect[2];
      dy = mpy - trimrect[3];
      d3 = sqrt(dx*dx + dy*dy);                                            //  SE

      dx = mpx - trimrect[0];
      dy = mpy - trimrect[3];
      d4 = sqrt(dx*dx + dy*dy);                                            //  SW

      Fcorner = 1;                                                         //  NW
      dd = d1;
      if (d2 < dd) { Fcorner = 2; dd = d2; }                               //  NE
      if (d3 < dd) { Fcorner = 3; dd = d3; }                               //  SE
      if (d4 < dd) { Fcorner = 4; dd = d4; }                               //  SW

      if (E3ww - mpx < dd && mpx > 0.9 * E3ww)                             //  mouse closer to right edge than corner
         if (mpy > 0.3 * E3hh && mpy < 0.7 * E3hh)                         //  if far from corners assume rotate
            goto mouse_rotate;                                             //  v.14.03

      if (Fcorner == 1) { trimrect[0] = mpx; trimrect[1] = mpy; }          //  move this corner to mouse
      if (Fcorner == 2) { trimrect[2] = mpx; trimrect[1] = mpy; }
      if (Fcorner == 3) { trimrect[2] = mpx; trimrect[3] = mpy; }
      if (Fcorner == 4) { trimrect[0] = mpx; trimrect[3] = mpy; }
   }

   if (trimrect[0] > trimrect[2]-10) trimrect[0] = trimrect[2]-10;         //  sanity limits
   if (trimrect[1] > trimrect[3]-10) trimrect[1] = trimrect[3]-10;

   zdialog_fetch(zd,"lock",rlock);                                         //  w/h ratio locked
   if (rlock && Fcorner) {
      if (Fcorner < 3)
         trimrect[3] = trimrect[1] + 1.0 * (trimrect[2] - trimrect[0]) / trimR;
      else
         trimrect[1] = trimrect[3] - 1.0 * (trimrect[2] - trimrect[0]) / trimR;
   }

   chop = 0;
   if (trimrect[0] < 0) {                                                  //  look for off the edge
      trimrect[0] = 0;
      chop = 1;
   }

   if (trimrect[2] > E3ww) {                                               //  after corner move
      trimrect[2] = E3ww;
      chop = 2;
   }

   if (trimrect[1] < 0) {
      trimrect[1] = 0;
      chop = 3;
   }

   if (trimrect[3] > E3hh) {
      trimrect[3] = E3hh;
      chop = 4;
   }

   if (rlock && chop) {                                                    //  keep ratio if off edge
      if (chop < 3)
         trimrect[3] = trimrect[1] + 1.0 * (trimrect[2] - trimrect[0]) / trimR;
      else
         trimrect[2] = trimrect[0] + 1.0 * (trimrect[3] - trimrect[1]) * trimR;
   }

   if (trimrect[0] > trimrect[2]-10) trimrect[0] = trimrect[2]-10;         //  sanity limits
   if (trimrect[1] > trimrect[3]-10) trimrect[1] = trimrect[3]-10;

   if (trimrect[0] < 0) trimrect[0] = 0;                                   //  keep within visible area
   if (trimrect[2] > E3ww) trimrect[2] = E3ww;
   if (trimrect[1] < 0) trimrect[1] = 0;
   if (trimrect[3] > E3hh) trimrect[3] = E3hh;

   trimsize[0] = FPR * (trimrect[2] - trimrect[0]);                        //  new rectangle dimensions
   trimsize[1] = FPR * (trimrect[3] - trimrect[1]);

   drr = 1.0 * trimsize[0] / trimsize[1];                                  //  new w/h ratio
   if (! rlock) trimR = drr;

   zdialog_stuff(zd,"width",trimsize[0]);                                  //  stuff width, height, ratio
   zdialog_stuff(zd,"height",trimsize[1]);
   sprintf(text,"%.2f  ",trimR);
   zdialog_stuff(zd,"ratio",text);

   trim_darkmargins();                                                     //  show trim area in image
   return;

// ------------------------------------------------------------------------

   mouse_rotate:

   Frotate = 1;
   Fcorner = 0;

   rotate_goal += 30.0 * ydrag / E3ww;                                     //  convert radians to degrees, reduced
   if (rotate_goal > 180) rotate_goal -= 360;                              //  keep within -180 to +180 deg.
   if (rotate_goal < -180) rotate_goal += 360;
   if (fabsf(rotate_goal) < 0.01) rotate_goal = 0;                         //  = 0 within my precision

   zdialog_stuff(zd,"degrees",rotate_goal);

   rotate_func();                                                          //  E3 is rotated E1
   return;
}


//  Darken image pixels outside of current trim margins.
//  messy logic: update pixmaps only for changed pixels to increase speed

void trimrotate_names::trim_darkmargins()
{
   using namespace trimrotate_names;

   int      ox1, oy1, ox2, oy2;                                            //  outer trim rectangle
   int      nx1, ny1, nx2, ny2;                                            //  inner trim rectangle
   int      px, py;
   float    *pix1, *pix3;

   if (trimrect[0] < 0) trimrect[0] = 0;                                   //  keep within visible area
   if (trimrect[2] > E3ww) trimrect[2] = E3ww;
   if (trimrect[1] < 0) trimrect[1] = 0;
   if (trimrect[3] > E3hh) trimrect[3] = E3hh;

   if (ptrimrect[0] < trimrect[0]) ox1 = ptrimrect[0];                     //  outer rectangle
   else ox1 = trimrect[0];
   if (ptrimrect[2] > trimrect[2]) ox2 = ptrimrect[2];
   else ox2 = trimrect[2];
   if (ptrimrect[1] < trimrect[1]) oy1 = ptrimrect[1];
   else oy1 = trimrect[1];
   if (ptrimrect[3] > trimrect[3]) oy2 = ptrimrect[3];
   else oy2 = trimrect[3];

   if (ptrimrect[0] > trimrect[0]) nx1 = ptrimrect[0];                     //  inner rectangle
   else nx1 = trimrect[0];
   if (ptrimrect[2] < trimrect[2]) nx2 = ptrimrect[2];
   else nx2 = trimrect[2];
   if (ptrimrect[1] > trimrect[1]) ny1 = ptrimrect[1];
   else ny1 = trimrect[1];
   if (ptrimrect[3] < trimrect[3]) ny2 = ptrimrect[3];
   else ny2 = trimrect[3];

   ox1 -= 2;                                                               //  expand outer rectangle
   oy1 -= 2;
   ox2 += 2;
   oy2 += 2;
   nx1 += 2;                                                               //  reduce inner rectangle
   ny1 += 2;
   nx2 -= 2;
   ny2 -= 2;

   if (ox1 < 0) ox1 = 0;
   if (oy1 < 0) oy1 = 0;
   if (ox2 > E3ww) ox2 = E3ww;
   if (oy2 > E3hh) oy2 = E3hh;

   if (nx1 < ox1) nx1 = ox1;
   if (ny1 < oy1) ny1 = oy1;
   if (nx2 > ox2) nx2 = ox2;
   if (ny2 > oy2) ny2 = oy2;

   if (nx2 < nx1) nx2 = nx1;
   if (ny2 < ny1) ny2 = ny1;

   for (py = oy1; py < ny1; py++)                                          //  top band of pixels
   for (px = ox1; px < ox2; px++)
   {
      pix1 = PXMpix(E9pxm,px,py);  
      pix3 = PXMpix(E3pxm,px,py);

      if (px < trimrect[0] || px > trimrect[2] || py < trimrect[1] || py > trimrect[3]) {
         pix3[0] = pix1[0] / 2;                                            //  outside trim margins
         pix3[1] = pix1[1] / 2;                                            //  50% brightness
         pix3[2] = pix1[2] / 2;
      }
      else {
         pix3[0] = pix1[0];                                                //  inside trim margins
         pix3[1] = pix1[1];                                                //  100% brightness
         pix3[2] = pix1[2];
      }
   }

   for (py = oy1; py < oy2; py++)                                          //  right band
   for (px = nx2; px < ox2; px++)
   {
      pix1 = PXMpix(E9pxm,px,py);
      pix3 = PXMpix(E3pxm,px,py);

      if (px < trimrect[0] || px > trimrect[2] || py < trimrect[1] || py > trimrect[3]) {
         pix3[0] = pix1[0] / 2;
         pix3[1] = pix1[1] / 2;
         pix3[2] = pix1[2] / 2;
      }
      else {
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }
   }

   for (py = ny2; py < oy2; py++)                                          //  bottom band
   for (px = ox1; px < ox2; px++)
   {
      pix1 = PXMpix(E9pxm,px,py);
      pix3 = PXMpix(E3pxm,px,py);

      if (px < trimrect[0] || px > trimrect[2] || py < trimrect[1] || py > trimrect[3]) {
         pix3[0] = pix1[0] / 2;
         pix3[1] = pix1[1] / 2;
         pix3[2] = pix1[2] / 2;
      }
      else {
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }
   }

   for (py = oy1; py < oy2; py++)                                          //  left band
   for (px = ox1; px < nx1; px++)
   {
      pix1 = PXMpix(E9pxm,px,py);
      pix3 = PXMpix(E3pxm,px,py);

      if (px < trimrect[0] || px > trimrect[2] || py < trimrect[1] || py > trimrect[3]) {
         pix3[0] = pix1[0] / 2;
         pix3[1] = pix1[1] / 2;
         pix3[2] = pix1[2] / 2;
      }
      else {
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }
   }

   mwpaint3(ox1,oy1,ox2-ox1,ny1-oy1);                                      //  4 updated rectangles
   mwpaint3(nx2,oy1,ox2-nx2,oy2-oy1);
   mwpaint3(ox1,ny2,ox2-ox1,oy2-ny2);
   mwpaint3(ox1,oy1,nx1-ox1,oy2-oy1);
   
   trimrotate_drawlines();                                                 //  draw trim rectangle etc.

   ptrimrect[0] = trimrect[0];                                             //  set prior trim rectangle
   ptrimrect[2] = trimrect[2];                                             //    from current trim rectangle
   ptrimrect[1] = trimrect[1];
   ptrimrect[3] = trimrect[3];

   return;
}


//  final trim - cut margins off the full size image
//  E3 is the input image, rotated and with black edges cropped if chosen
//  E3 is the output image from user trim rectangle
//  FPR should be 1.0 and E3 should be full size

void trimrotate_names::trim_final()
{
   using namespace trimrotate_names;

   int      px1, py1, px2, py2;
   float    *pix3, *pix9;

   draw_toplines(2);                                                       //  erase trim rectangle
   erase_topcircles();                                                     //    and target circle

   if (trimrect[2] > E3ww) trimrect[2] = E3ww;                             //  final check
   if (trimrect[3] > E3hh) trimrect[3] = E3hh;

   trimsize[0] = trimrect[2] - trimrect[0];                                //  new rectangle dimensions
   trimsize[1] = trimrect[3] - trimrect[1];

   PXM_free(E9pxm);
   E9pxm = PXM_make(trimsize[0],trimsize[1]);                              //  new pixmap with requested size

   for (py1 = trimrect[1]; py1 < trimrect[3]; py1++)                       //  copy E3 (rotated) to new size
   for (px1 = trimrect[0]; px1 < trimrect[2]; px1++)
   {
      px2 = px1 - trimrect[0];
      py2 = py1 - trimrect[1];
      pix3 = PXMpix(E3pxm,px1,py1);
      pix9 = PXMpix(E9pxm,px2,py2);
      pix9[0] = pix3[0];
      pix9[1] = pix3[1];
      pix9[2] = pix3[2];
   }

   PXM_free(E3pxm);
   E3pxm = E9pxm;
   E9pxm = 0;

   E3ww = E3pxm->ww;                                                       //  update E3 dimensions
   E3hh = E3pxm->hh;
   FPR = 1.0 * (E0ww + E0hh) / (E3ww + E3hh);                              //  full image / preview ratio

   CEF->Fmods++;
   CEF->Fsaved = 0;
   
   ptrimsize[0] = trimsize[0] + 0.5;                                       //  save final trim size for poss.     v.14.01
   ptrimsize[1] = trimsize[1] + 0.5;                                       //    recall via [Previous] button

   mwpaint2();
   return;
}


//  rotate function
//  E3 = E1 with rotation and edges cropped if chosen.
//  E9 = E3 copy: used as source for darkened pixels in E3.

void trimrotate_names::rotate_func()
{
   using namespace trimrotate_names;

   int         px3, py3, px9, py9;
   int         wwcut, hhcut, ww, hh;
   float       trim_angle, radians;
   float       *pix3, *pix9;
   zdialog     *zd = EFtrimrotate.zd;

   zmainloop();                                                            //  updates before big CPU time     v.13.12

   if (fabs(rotate_goal) < 0.01) {                                         //  no rotation
      rotate_goal = rotate_angle = 0.0;
      zdialog_stuff(zd,"degrees",0.0);
      PXM_free(E3pxm);                                                     //  E1 >> E3
      E3pxm = PXM_copy(E1pxm);
      PXM_free(E9pxm);                                                     //  E1 >> E9
      E9pxm = PXM_copy(E1pxm);
      CEF->Fmods = 0;                                                      //  will be overridden by trim
   }

   if (rotate_goal != rotate_angle || rotatecrop2 != rotatecrop1)
   {
      rotate_angle = rotate_goal;
      rotatecrop2 = rotatecrop1;                                           //  remember autocrop setting

      PXM_free(E9pxm);
      E9pxm = PXM_rotate(E1pxm,rotate_angle);                              //  E9 is rotated E1

      if (rotatecrop1)                                                     //  edge cropping is chosen
      {
         trim_angle = fabsf(rotate_goal);
         while (trim_angle > 45) trim_angle -= 90;
         radians = fabsf(trim_angle / 57.296);
         wwcut = int(E9pxm->hh * sinf(radians) + 1);                       //  amount to crop off edges
         hhcut = int(E9pxm->ww * sinf(radians) + 1);

         if (wwcut > E9pxm->ww / 2 - 10)                                   //  do not cut 100%
            wwcut = E9pxm->ww / 2 - 10;
         if (hhcut > E9pxm->hh / 2 - 10)
            hhcut = E9pxm->hh / 2 - 10;

         ww = E9pxm->ww - 2 * wwcut;
         hh = E9pxm->hh - 2 * hhcut;

         PXM_free(E3pxm);
         E3pxm = PXM_make(ww,hh);                                          //  E3 = cropped E9

         for (py3 = 0; py3 < hh; py3++)                                    //  copy pixels E9 > E3
         for (px3 = 0; px3 < ww; px3++)
         {
            pix3 = PXMpix(E3pxm,px3,py3);
            px9 = px3 + wwcut;
            py9 = py3 + hhcut;
            pix9 = PXMpix(E9pxm,px9,py9);
            pix3[0] = pix9[0];
            pix3[1] = pix9[1];
            pix3[2] = pix9[2];
         }

         PXM_free(E9pxm);                                                  //  E9 = E3
         E9pxm = PXM_copy(E3pxm);
      }

      else {                                                               //  no edge cropping
         PXM_free(E3pxm);                                                  //  E3 = E9
         E3pxm = PXM_copy(E9pxm);
      }
   }

   E3ww = E3pxm->ww;                                                       //  update E3 dimensions
   E3hh = E3pxm->hh;

   if (trimrect[2] > E3ww) trimrect[2] = E3ww;                             //  contract trim rectangle if needed
   if (trimrect[3] > E3hh) trimrect[3] = E3hh;

   trimsize[0] = FPR * (trimrect[2] - trimrect[0]);                        //  new rectangle dimensions
   trimsize[1] = FPR * (trimrect[3] - trimrect[1]);

   zdialog_stuff(zd,"width",trimsize[0]);                                  //  stuff width, height
   zdialog_stuff(zd,"height",trimsize[1]);

   mwpaintnow();
   trimrotate_drawlines();                                                 //  draw trim rectangle etc.

   ptrimrect[0] = 0;                                                       //  set prior trim rectangle
   ptrimrect[1] = 0;
   ptrimrect[2] = E3ww;                                                    //    = 100% of image
   ptrimrect[3] = E3hh;

   return;
}


//  perform final rotate and trim on full-size image

void trimrotate_names::trimrotate_final()
{
   using namespace trimrotate_names;

   if (CEF->Fpreview)                                                      //  preview image was used
   {
      CEF->Fpreview = 0;

      PXM_free(E1pxm);                                                     //  free preview images
      PXM_free(E3pxm);
      PXM_free(E9pxm);

      E1pxm = PXM_copy(E0pxm);                                             //  E0 >> E1, full size image

      trimrect[0] = FPR * trimrect[0];                                     //  full scale trim rectangle
      trimrect[1] = FPR * trimrect[1];
      trimrect[2] = FPR * trimrect[2];
      trimrect[3] = FPR * trimrect[3];
      FPR = 1.0;

      trimsize[0] = trimrect[2] - trimrect[0];                             //  final trim size
      trimsize[1] = trimrect[3] - trimrect[1];

      rotate_angle = 0;                                                    //  rotate full size image
      rotate_func();                                                       //  E3 = E9 = rotated and edge-cropped E1
   }

   trim_final();                                                           //  trim E3 from user trim rectangle
   return;
}


//  draw lines on image: trim rectangle, guidelines, gridlines

void trimrotate_names::trimrotate_drawlines()
{
   using namespace trimrotate_names;

   int      px, py, radius;

   Ntoplines = 4;                                                          //  outline trim rectangle
   toplinex1[0] = trimrect[0];
   topliney1[0] = trimrect[1];
   toplinex2[0] = trimrect[2];
   topliney2[0] = trimrect[1];
   toplinex1[1] = trimrect[2];
   topliney1[1] = trimrect[1];
   toplinex2[1] = trimrect[2];
   topliney2[1] = trimrect[3];
   toplinex1[2] = trimrect[2];
   topliney1[2] = trimrect[3];
   toplinex2[2] = trimrect[0];
   topliney2[2] = trimrect[3];
   toplinex1[3] = trimrect[0];
   topliney1[3] = trimrect[3];
   toplinex2[3] = trimrect[0];
   topliney2[3] = trimrect[1];

   if (Fguidelines) {                                                      //  vert/horz rotate guidelines
      Ntoplines = 6;
      toplinex1[4] = guidelineX;
      topliney1[4] = 0;
      toplinex2[4] = guidelineX;
      topliney2[4] = E3hh-1;
      toplinex1[5] = 0;
      topliney1[5] = guidelineY;
      toplinex2[5] = E3ww-1;
      topliney2[5] = guidelineY;
   }

   draw_toplines(1);                                                       //  draw trim rectangle and guidelines

   erase_topcircles();
   px = 0.5 * (ptrimrect[0] + ptrimrect[2]);
   py = 0.5 * (ptrimrect[1] + ptrimrect[3]);
   radius = 0.02 * Mscale * (E3ww + E3hh);
   mwpaint3(px-radius,py-radius,2*radius,2*radius);                        //  bugfix                             v.14.01

   px = 0.5 * (trimrect[0] + trimrect[2]);
   py = 0.5 * (trimrect[1] + trimrect[3]);
   add_topcircle(px,py,radius/2,1);
   add_topcircle(px,py,radius/2+1,2);
   draw_topcircles();
   
   if (gridsettings[currgrid][GON]) mwpaint2();                            //  refresh gridlines, delayed         v.14.01

   return;
}


/**************************************************************************/

//  auto-trim image - set trim rectangle to exclude black margins
//                    left over from rotate or warp functions

void m_autotrim(GtkWidget *, cchar *)
{
   int      px1, py1, px2, py2;
   int      qx1, qy1, qx2, qy2;
   int      qx, qy, step1, step2;
   int      area1, area2, Fgrow = 0;
   float    *ppix;

   if (! curr_file) return;
   if (! E0pxm) {
      E0pxm = PXM_load(curr_file,1);                                       //  create E0 image for my use      v.13.11
      if (! E0pxm) return;
   }

   if (checkpend("edit lock")) return;                                     //  v.13.12
   Fmenulock = 1;
   Ffuncbusy++;  

   px1 = 0.4 * Fpxb->ww;                                                   //  select small rectangle in the middle
   py1 = 0.4 * Fpxb->hh;
   px2 = 0.6 * Fpxb->ww;
   py2 = 0.6 * Fpxb->hh;

   step1 = 0.02 * (Fpxb->ww + Fpxb->hh);                                   //  start with big search steps
   step2 = 0.2 * step1;                                                    //  v.13.11
   if (step2 < 1) step2 = 1;

   while (true)
   {
      while (true)
      {
         Fgrow = 0;

         area1 = (px2 - px1) * (py2 - py1);                                //  area of current selection rectangle
         area2 = area1;

         for (qx1 = px1-step1; qx1 <= px1+step1; qx1 += step2)             //  loop, vary NW and SE corners +/-
         for (qy1 = py1-step1; qy1 <= py1+step1; qy1 += step2)
         for (qx2 = px2-step1; qx2 <= px2+step1; qx2 += step2)
         for (qy2 = py2-step1; qy2 <= py2+step1; qy2 += step2)
         {
            if (qx1 < 0) continue;                                         //  check image limits
            if (qy1 < 0) continue;
            if (qx1 > 0.5 * Fpxb->ww) continue;
            if (qy1 > 0.5 * Fpxb->hh) continue;
            if (qx2 > Fpxb->ww-1) continue;
            if (qy2 > Fpxb->hh-1) continue;
            if (qx2 < 0.5 * Fpxb->ww) continue;
            if (qy2 < 0.5 * Fpxb->hh) continue;

            ppix = PXMpix(E0pxm,qx1,qy1);                                  //  check 4 corners are not
            if (ppix[0] < 1 && ppix[1] < 1 && ppix[2] < 1) continue;       //    in the black margin zones
            ppix = PXMpix(E0pxm,qx2,qy1);
            if (ppix[0] < 1 && ppix[1] < 1 && ppix[2] < 1) continue;       //  check all colors             v.13.11
            ppix = PXMpix(E0pxm,qx2,qy2);
            if (ppix[0] < 1 && ppix[1] < 1 && ppix[2] < 1) continue;
            ppix = PXMpix(E0pxm,qx1,qy2);
            if (ppix[0] < 1 && ppix[1] < 1 && ppix[2] < 1) continue;

            area2 = (qx2 - qx1) * (qy2 - qy1);                             //  look for larger enclosed area
            if (area2 <= area1) continue;

            for (qx = qx1; qx < qx2; qx++) {                               //  check top/bottom sides don't intersect
               ppix = PXMpix(E0pxm,qx,qy1);                                //    the black margin zones
               if (ppix[0] < 1 && ppix[1] < 1 && ppix[2] < 1) break;
               ppix = PXMpix(E0pxm,qx,qy2);
               if (ppix[0] < 1 && ppix[1] < 1 && ppix[2] < 1) break;
            }
            if (qx < qx2) continue;

            for (qy = qy1; qy < qy2; qy++) {                               //  also left/right sides
               ppix = PXMpix(E0pxm,qx1,qy);
               if (ppix[0] < 1 && ppix[1] < 1 && ppix[2] < 1) break;
               ppix = PXMpix(E0pxm,qx2,qy);
               if (ppix[0] < 1 && ppix[1] < 1 && ppix[2] < 1) break;
            }
            if (qy < qy2) continue;

            Fgrow = 1;                                                     //  successfully grew the rectangle
            px1 = qx1;                                                     //  new bigger rectangle coordinates
            py1 = qy1;                                                     //    for the next search iteration
            px2 = qx2;
            py2 = qy2;
            goto breakout;                                                 //  leave all 4 loops      v.13.08
         }

      breakout:
         if (! Fgrow) break;
         zmainloop();                                                      //  v.13.11
      }

      if (step2 == 1) break;                                               //  done

      step1 = 0.6 * step1;                                                 //  reduce search step size
      if (step1 < 2) step1 = 2;
      step2 = 0.2 * step1;
      if (step2 < 1) step2 = 1;
   }

   trimrect[0] = px1;                                                      //  set parameters for trim function
   trimrect[2] = px2;
   trimrect[1] = py1;
   trimrect[3] = py2;

   Fmenulock = 0;                                                          //  v.13.12 
   Ffuncbusy--;  

   m_trimrotate(0,"autotrim");                                             //  perform regular trim with
   return;                                                                 //    pre-set margins
}


/**************************************************************************/

//  Resize (rescale) image
//
//  Output pixels are composites of input pixels, e.g. 2/3 size means
//  that 3x3 input pixels are mapped into 2x2 output pixels, and an
//  image size of 1000 x 600 becomes 667 x 400.

int      resize_ww0, resize_hh0;                                           //  original size
int      resize_ww1, resize_hh1;                                           //  new size
editfunc    EFresize;

void   resize_initz(zdialog *zd);
int    resize_dialog_event(zdialog *zd, cchar *event);
void * resize_thread(void *);


void m_resize(GtkWidget *, cchar *)
{
   cchar  *lockmess = ZTX("Lock aspect ratio");

   F1_help_topic = "resize";

   EFresize.menufunc = m_resize;
   EFresize.funcname = "resize";
   EFresize.Frestart = 1;                                                  //  allow restart             v.13.04
   EFresize.threadfunc = resize_thread;                                    //  thread function
   if (! edit_setup(EFresize)) return;                                     //  setup edit

/****
                  pixels     percent
          width   [______]   [______]
          height  [______]   [______]

          presets [2/3] [1/2] [1/3] [1/4] [Prev]

          [_] lock aspect ratio

                            [done] [cancel]
****/

   zdialog *zd = zdialog_new(ZTX("Resize Image"),mWin,Bdone,Bcancel,null);
   EFresize.zd = zd;
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"vbox","vb11","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb12","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb13","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"label","placeholder","vb11",0);
   zdialog_add_widget(zd,"label","labw","vb11",Bwidth);
   zdialog_add_widget(zd,"label","labh","vb11",Bheight);
   zdialog_add_widget(zd,"label","labpix","vb12","pixels");
   zdialog_add_widget(zd,"spin","wpix","vb12","20|9999|1|20");
   zdialog_add_widget(zd,"spin","hpix","vb12","20|9999|1|20");
   zdialog_add_widget(zd,"label","labpct","vb13",Bpercent);
   zdialog_add_widget(zd,"spin","wpct","vb13","1|500|0.1|100");
   zdialog_add_widget(zd,"spin","hpct","vb13","1|500|0.1|100");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","preset","hb2",Bpresets,"space=5");
   zdialog_add_widget(zd,"button","b 3/4","hb2","3/4");
   zdialog_add_widget(zd,"button","b 2/3","hb2","2/3");
   zdialog_add_widget(zd,"button","b 1/2","hb2","1/2");
   zdialog_add_widget(zd,"button","b 1/3","hb2","1/3");
   zdialog_add_widget(zd,"button","b 1/4","hb2","1/4");
   zdialog_add_widget(zd,"button","prev","hb2",Bprev);
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"check","lock","hb3",lockmess,"space=5");

   resize_initz(zd);
   zdialog_run(zd,resize_dialog_event,"save");                             //  run dialog

   return;
}


//  initialization for new image

void resize_initz(zdialog *zd)
{
   resize_ww0 = E1pxm->ww;                                                 //  original width, height
   resize_hh0 = E1pxm->hh;
   zdialog_stuff(zd,"wpix",resize_ww0);
   zdialog_stuff(zd,"hpix",resize_hh0);
   zdialog_stuff(zd,"lock",1);
   return;
}


//  dialog event and completion callback function

int resize_dialog_event(zdialog *zd, cchar *event)
{
   int         lock;
   float       wpct1, hpct1;

   if (strEqu(event,"done")) zd->zstat = 1;                                //  apply and quit               v.13.12
   if (strEqu(event,"enter")) zd->zstat = 1;                               //  v.14.03

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) {                                                //  done
         editresize[0] = resize_ww1;                                       //  remember size used
         editresize[1] = resize_hh1;
         edit_done(0);
      }
      else edit_cancel(0);                                                 //  cancel or kill

      Fzoom = 0;
      return 1;
   }

   if (strEqu(event,"focus")) return 1;                                    //  ignore focus                 v.13.09

   zdialog_fetch(zd,"wpix",resize_ww1);                                    //  get all widget values
   zdialog_fetch(zd,"hpix",resize_hh1);
   zdialog_fetch(zd,"wpct",wpct1);
   zdialog_fetch(zd,"hpct",hpct1);
   zdialog_fetch(zd,"lock",lock);

   if (strEqu(event,"b 3/4")) {
      resize_ww1 = (3 * resize_ww0 + 3) / 4;
      resize_hh1 = (3 * resize_hh0 + 3) / 4;
   }

   if (strEqu(event,"b 2/3")) {
      resize_ww1 = (2 * resize_ww0 + 2) / 3;
      resize_hh1 = (2 * resize_hh0 + 2) / 3;
   }

   if (strEqu(event,"b 1/2")) {
      resize_ww1 = (resize_ww0 + 1) / 2;
      resize_hh1 = (resize_hh0 + 1) / 2;
   }

   if (strEqu(event,"b 1/3")) {
      resize_ww1 = (resize_ww0 + 2) / 3;
      resize_hh1 = (resize_hh0 + 2) / 3;
   }

   if (strEqu(event,"b 1/4")) {
      resize_ww1 = (resize_ww0 + 3) / 4;
      resize_hh1 = (resize_hh0 + 3) / 4;
   }

   if (strEqu(event,"prev")) {                                             //  use previous resize values
      resize_ww1 = editresize[0];
      resize_hh1 = editresize[1];
   }

   if (strEqu(event,"wpct"))                                               //  width % - set pixel width
      resize_ww1 = int(wpct1 / 100.0 * resize_ww0 + 0.5);

   if (strEqu(event,"hpct"))                                               //  height % - set pixel height
      resize_hh1 = int(hpct1 / 100.0 * resize_hh0 + 0.5);

   if (lock && event[0] == 'w')                                            //  preserve width/height ratio
      resize_hh1 = int(resize_ww1 * (1.0 * resize_hh0 / resize_ww0) + 0.5);
   if (lock && event[0] == 'h')
      resize_ww1 = int(resize_hh1 * (1.0 * resize_ww0 / resize_hh0) + 0.5);

   if (resize_ww1 < 20 || resize_hh1 < 20) return 1;                       //  refuse tiny size       v.12.11

   hpct1 = 100.0 * resize_hh1 / resize_hh0;                                //  set percents to match pixels
   wpct1 = 100.0 * resize_ww1 / resize_ww0;

   zdialog_stuff(zd,"wpix",resize_ww1);                                    //  index all widget values
   zdialog_stuff(zd,"hpix",resize_hh1);
   zdialog_stuff(zd,"wpct",wpct1);
   zdialog_stuff(zd,"hpct",hpct1);

   signal_thread();                                                        //  update image, no wait for idle

   return 1;
}


//  do the resize job

void * resize_thread(void *)
{
   while (true)
   {
      thread_idle_loop();                                                  //  wait for signal

      mutex_lock(&Fpixmap_lock);
      PXM_free(E3pxm);
      E3pxm = PXM_rescale(E1pxm,resize_ww1,resize_hh1);                    //  rescale the edit image
      if (! E3pxm) thread_exit();                                          //  memory failure      v.13.03
      CEF->Fmods++;
      CEF->Fsaved = 0;
      mutex_unlock(&Fpixmap_lock);
      mwpaint2();
      zsleep(0.1);                                                         //  v.13.03
   }

   return 0;
}


/**************************************************************************/

//  flip an image horizontally or vertically

editfunc    EFflip;

void m_flip(GtkWidget *, cchar *)
{
   int flip_dialog_event(zdialog *zd, cchar *event);

   F1_help_topic = "flip_image";

   EFflip.menufunc = m_flip;
   EFflip.funcname = "flip";
   EFflip.Frestart = 1;                                                    //  allow restart             v.13.04
   if (! edit_setup(EFflip)) return;                                       //  setup edit

   zdialog *zd = zdialog_new(ZTX("Flip"),mWin,Bdone,Bcancel,null);
   EFflip.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"button","horz","hb1",ZTX("horizontal"),"space=5");
   zdialog_add_widget(zd,"button","vert","hb1",ZTX("vertical"),"space=5");

   zdialog_run(zd,flip_dialog_event,"save");                               //  run dialog, parallel
   return;
}


//  dialog event and completion callback function

int flip_dialog_event(zdialog *zd, cchar *event)
{
   int flip_horz();
   int flip_vert();

   if (strEqu(event,"done")) zd->zstat = 1;                                //  apply and quit               v.13.12
   if (strEqu(event,"enter")) zd->zstat = 1;                               //  v.14.03

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done(0);
      else edit_cancel(0);
      return 1;
   }

   if (strEqu(event,"horz")) flip_horz();
   if (strEqu(event,"vert")) flip_vert();
   return 1;
}


int flip_horz()
{
   int      px, py;
   float    *pix3, *pix9;

   E9pxm = PXM_copy(E3pxm);

   for (py = 0; py < E3pxm->hh; py++)
   for (px = 0; px < E3pxm->ww; px++)
   {
      pix3 = PXMpix(E3pxm,px,py);                                          //  image9 = flipped image3
      pix9 = PXMpix(E9pxm,E3pxm->ww-1-px,py);
      pix9[0] = pix3[0];
      pix9[1] = pix3[1];
      pix9[2] = pix3[2];
   }

   mutex_lock(&Fpixmap_lock);
   PXM_free(E3pxm);                                                        //  image9 >> image3
   E3pxm = E9pxm;
   E9pxm = 0;
   mutex_unlock(&Fpixmap_lock);

   CEF->Fmods++;
   CEF->Fsaved = 0;
   mwpaint2();
   return 0;
}


int flip_vert()
{
   int      px, py;
   float    *pix3, *pix9;

   E9pxm = PXM_copy(E3pxm);

   for (py = 0; py < E3pxm->hh; py++)
   for (px = 0; px < E3pxm->ww; px++)
   {
      pix3 = PXMpix(E3pxm,px,py);                                          //  image9 = flipped image3
      pix9 = PXMpix(E9pxm,px,E3pxm->hh-1-py);
      pix9[0] = pix3[0];
      pix9[1] = pix3[1];
      pix9[2] = pix3[2];
   }

   mutex_lock(&Fpixmap_lock);
   PXM_free(E3pxm);                                                        //  image9 >> image3
   E3pxm = E9pxm;
   E9pxm = 0;
   mutex_unlock(&Fpixmap_lock);

   CEF->Fmods++;
   CEF->Fsaved = 0;
   mwpaint2();
   return 0;
}


/**************************************************************************/

//  write text on top of the image

namespace writetext
{
   textattr_t  attr;

   char     text[1000] = "text";                                           //  text
   char     file[1000] = "";                                               //  file for write_text data
   int      px, py;                                                        //  text position on image
   int      present;                                                       //  flag, write_text present on image
   PXM      *pxm_text = 0;                                                 //  buffer for rendered text
   PXM      *pxm_transp = 0;                                               //  buffer for transparency data

   int   dialog_event(zdialog *zd, cchar *event);                          //  dialog event function
   void  mousefunc();                                                      //  mouse event function
   void  write(int mode);                                                  //  write text on image

   editfunc    EFwritetext;
}


void m_write_text(GtkWidget *, cchar *menu)
{
   using namespace writetext;

   cchar    *intro = ZTX("Enter text, click/drag on image, right click to remove");

   F1_help_topic = "write_text";                                           //  user guide topic

   EFwritetext.menufunc = m_write_text;
   EFwritetext.funcname = "write_text";
   EFwritetext.Farea = 1;                                                  //  select area ignored
   EFwritetext.Frestart = 1;                                               //  allow restart                v.13.05
   if (! edit_setup(EFwritetext)) return;                                  //  setup edit

/***
       ____________________________________________________________
      |               Write Text on Image                          |
      |                                                            |
      |  Enter text, click/drag on image, right click to remove.   |
      |                                                            |
      |  Text [__________________________________________________] |
      |                                                            |
      |  [font] [FreeSans         ]  size [ 44|v]                  |
      |                                                            |
      |            color   transp.   width    angle                |
      |  text     [_____]  [___|v]           [___|v]               |
      |  backing  [_____]  [___|v]                                 |
      |  outline  [_____]  [___|v]  [___|v]                        |
      |  shadow   [_____]  [___|v]  [___|v]  [___|v]               |
      |                                                            |
      |  Text File: [Open] [Save]    Metadata: [Caption] [Comment] |
      |                                                            |
      |                                    [Apply] [Done] [Cancel] |
      |____________________________________________________________|

***/

   zdialog *zd = zdialog_new(ZTX("Write Text on Image"),mWin,Bapply,Bdone,Bcancel,null);
   EFwritetext.zd = zd;
   EFwritetext.mousefunc = mousefunc;
   EFwritetext.menufunc = m_write_text;                                    //  allow restart          v.13.04

   zdialog_add_widget(zd,"label","intro","dialog",intro,"space=3");
   zdialog_add_widget(zd,"hbox","hbtext","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labtext","hbtext",ZTX("Text"),"space=5");
   zdialog_add_widget(zd,"frame","frtext","hbtext",0,"expand");
   zdialog_add_widget(zd,"edit","text","frtext","text","expand|wrap");     //  v.13.12

   zdialog_add_widget(zd,"hbox","hbfont","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","fontbutt","hbfont",Bfont);
   zdialog_add_widget(zd,"entry","fontname","hbfont","FreeSans","space=2|expand");
   zdialog_add_widget(zd,"label","labsize","hbfont",Bsize,"space=2");
   zdialog_add_widget(zd,"spin","fontsize","hbfont","8|500|1|40");

   zdialog_add_widget(zd,"hbox","hbcol","dialog");
   zdialog_add_widget(zd,"vbox","vbcol1","hbcol",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vbcol2","hbcol",0,"homog|space=2");
   zdialog_add_widget(zd,"vbox","vbcol3","hbcol",0,"homog|space=2");
   zdialog_add_widget(zd,"vbox","vbcol4","hbcol",0,"homog|space=2");
   zdialog_add_widget(zd,"vbox","vbcol5","hbcol",0,"homog|space=2");

   zdialog_add_widget(zd,"label","space","vbcol1");
   zdialog_add_widget(zd,"label","labtext","vbcol1",ZTX("text"));
   zdialog_add_widget(zd,"label","labback","vbcol1",ZTX("backing"));
   zdialog_add_widget(zd,"label","laboutln","vbcol1",ZTX("outline"));
   zdialog_add_widget(zd,"label","labshadow","vbcol1",ZTX("shadow"));

   zdialog_add_widget(zd,"label","labcol","vbcol2",ZTX("Color"));
   zdialog_add_widget(zd,"colorbutt","fgcolor","vbcol2","0|0|0");
   zdialog_add_widget(zd,"colorbutt","bgcolor","vbcol2","255|255|255");
   zdialog_add_widget(zd,"colorbutt","tocolor","vbcol2","255|0|0");
   zdialog_add_widget(zd,"colorbutt","shcolor","vbcol2","255|0|0");

   zdialog_add_widget(zd,"label","labcol","vbcol3",ZTX("Transparency"));
   zdialog_add_widget(zd,"spin","fgtransp","vbcol3","0|100|1|0");
   zdialog_add_widget(zd,"spin","bgtransp","vbcol3","0|100|1|0");
   zdialog_add_widget(zd,"spin","totransp","vbcol3","0|100|1|0");
   zdialog_add_widget(zd,"spin","shtransp","vbcol3","0|100|1|0");

   zdialog_add_widget(zd,"label","labw","vbcol4",ZTX("Width"));
   zdialog_add_widget(zd,"label","space","vbcol4");
   zdialog_add_widget(zd,"label","space","vbcol4");
   zdialog_add_widget(zd,"spin","towidth","vbcol4","0|30|1|0");
   zdialog_add_widget(zd,"spin","shwidth","vbcol4","0|50|1|0");

   zdialog_add_widget(zd,"label","labw","vbcol5",Bangle);
   zdialog_add_widget(zd,"spin","angle","vbcol5","-180|180|0.5|0");
   zdialog_add_widget(zd,"label","space","vbcol5");
   zdialog_add_widget(zd,"label","space","vbcol5");
   zdialog_add_widget(zd,"spin","shangle","vbcol5","-180|180|1|0");

   zdialog_add_widget(zd,"hbox","hbtext","dialog",0,"space=8");
   zdialog_add_widget(zd,"label","labtf","hbtext",ZTX("Text File:"),"space=3");
   zdialog_add_widget(zd,"button","loadtext","hbtext",Bopen);
   zdialog_add_widget(zd,"button","savetext","hbtext",Bsave);

   zdialog_add_widget(zd,"label","space","hbtext",0,"space=3");            //  v.13.12
   zdialog_add_widget(zd,"label","labmeta","hbtext",ZTX("Metadata:"),"space=3");
   zdialog_add_widget(zd,"button","caption","hbtext",ZTX("Caption"));
   zdialog_add_widget(zd,"button","comment","hbtext",ZTX("Comment"));
   
   zdialog_restore_inputs(zd);                                             //  restore prior inputs
   zdialog_run(zd,dialog_event,"save");                                    //  run dialog, parallel

   zdialog_fetch(zd,"text",text,1000);                                     //  get defaults or prior inputs
   zdialog_fetch(zd,"fontname",attr.font,80);
   zdialog_fetch(zd,"fontsize",attr.size);
   zdialog_fetch(zd,"angle",attr.angle);
   zdialog_fetch(zd,"fgcolor",attr.color[0],20);
   zdialog_fetch(zd,"bgcolor",attr.color[1],20);
   zdialog_fetch(zd,"tocolor",attr.color[2],20);
   zdialog_fetch(zd,"shcolor",attr.color[3],20);
   zdialog_fetch(zd,"fgtransp",attr.transp[0]);
   zdialog_fetch(zd,"bgtransp",attr.transp[1]);
   zdialog_fetch(zd,"totransp",attr.transp[2]);
   zdialog_fetch(zd,"shtransp",attr.transp[3]);
   zdialog_fetch(zd,"towidth",attr.towidth);
   zdialog_fetch(zd,"shwidth",attr.shwidth);
   zdialog_fetch(zd,"shangle",attr.shangle);

   attr.text = text;                                                       //  v.13.10
   gentext(&attr,pxm_text,pxm_transp);                                     //  initial text

   takeMouse(mousefunc,dragcursor);                                        //  connect mouse function
   present = 0;                                                            //  no text on image yet
   px = py = -1;                                                           //  no position defined yet

   return;
}


//  dialog event and completion callback function

int writetext::dialog_event(zdialog *zd, cchar *event)
{
   using namespace writetext;

   GtkWidget   *font_dialog;
   char        font[100];                                                  //  font name and size
   int         size;
   char        *pp;
   cchar       *keyname;
   char        **keyvals;

   if (strEqu(event,"done")) zd->zstat = 2;                                //  apply and quit               v.13.12

   if (zd->zstat)
   {
      if (zd->zstat == 1) {                                                //  Apply, commit present text to image
         write(1);                                                         //  v.12.03
         present = 0;                                                      //  (no old text to erase)
         px = py = -1;
         zd->zstat = 0;
         edit_done(0);                                                     //  finish
         PXM_free(pxm_text);                                               //  free memory                  v.12.10
         PXM_free(pxm_transp);
         m_write_text(0,0);                                                //  start again
         return 1;
      }

      if (zd->zstat == 2 && CEF->Fmods) edit_done(0);                      //  Done, complete pending edit
      else edit_cancel(0);                                                 //  Cancel or kill
      PXM_free(pxm_text);                                                  //  free memory                  v.12.10
      PXM_free(pxm_transp);
      return 1;
   }

   if (strEqu(event,"focus")) {                                            //  toggle mouse capture
      takeMouse(mousefunc,dragcursor);                                     //  connect mouse function
      return 1;
   }

   if (strEqu(event,"loadtext"))                                           //  load zdialog fields from file
   {
      load_text(zd);
      zdialog_fetch(zd,"text",text,1000);                                  //  get all zdialog fields
      zdialog_fetch(zd,"fontname",attr.font,80);
      zdialog_fetch(zd,"fontsize",attr.size);
      zdialog_fetch(zd,"angle",attr.angle);
      zdialog_fetch(zd,"fgcolor",attr.color[0],20);
      zdialog_fetch(zd,"bgcolor",attr.color[1],20);
      zdialog_fetch(zd,"tocolor",attr.color[2],20);
      zdialog_fetch(zd,"shcolor",attr.color[3],20);
      zdialog_fetch(zd,"fgtransp",attr.transp[0]);
      zdialog_fetch(zd,"bgtransp",attr.transp[1]);
      zdialog_fetch(zd,"totransp",attr.transp[2]);
      zdialog_fetch(zd,"shtransp",attr.transp[3]);
      zdialog_fetch(zd,"towidth",attr.towidth);
      zdialog_fetch(zd,"shwidth",attr.shwidth);
      zdialog_fetch(zd,"shangle",attr.shangle);
   }

   if (strEqu(event,"savetext")) {                                         //  save zdialog fields to file
      save_text(zd);
      return 1;
   }
   
   if (strstr("caption comment",event)) {                                  //  get caption or comment metadata    v.13.12
      if (strEqu(event,"caption")) keyname = iptc_caption_key;
      else keyname = exif_comment_key;
      keyvals = exif_get(curr_file,&keyname,1);
      if (! *keyvals) return 1;
      if (strlen(*keyvals) > 999) *keyvals[999] = 0;
      repl_1str(*keyvals,text,"\\n","\n");                                 //  replace "\n" with newlines
      free(*keyvals);
      zdialog_stuff(zd,"text",text);                                       //  stuff dialog with metadata
   }

   if (strEqu(event,"text"))                                               //  get text from dialog
      zdialog_fetch(zd,"text",text,1000);

   if (strEqu(event,"fontbutt")) {                                         //  select new font
      snprintf(font,100,"%s %d",attr.font,attr.size);
      font_dialog = gtk_font_chooser_dialog_new(ZTX("select font"),MWIN);
      gtk_font_chooser_set_font(GTK_FONT_CHOOSER(font_dialog),font);
      gtk_dialog_run(GTK_DIALOG(font_dialog));
      pp = gtk_font_chooser_get_font(GTK_FONT_CHOOSER(font_dialog));
      gtk_widget_destroy(font_dialog);

      if (pp) {                                                            //  should have "fontname nn"
         strncpy0(font,pp,100);
         g_free(pp);
         pp = font + strlen(font);
         while (*pp != ' ') pp--;
         if (pp > font) {
            size = atoi(pp);
            if (size < 8) size = 8;
            zdialog_stuff(zd,"fontsize",size);
            attr.size = size;
            *pp = 0;
            strncpy0(attr.font,font,80);                                   //  get fontname = new font name
            zdialog_stuff(zd,"fontname",font);
         }
      }
   }

   if (strEqu(event,"fontsize"))                                           //  new font size
      zdialog_fetch(zd,"fontsize",attr.size);

   if (strEqu(event,"angle"))
      zdialog_fetch(zd,"angle",attr.angle);

   if (strEqu(event,"fgcolor"))                                            //  foreground (text) color
      zdialog_fetch(zd,"fgcolor",attr.color[0],20);

   if (strEqu(event,"bgcolor"))                                            //  background color
      zdialog_fetch(zd,"bgcolor",attr.color[1],20);

   if (strEqu(event,"tocolor"))                                            //  text outline color
      zdialog_fetch(zd,"tocolor",attr.color[2],20);

   if (strEqu(event,"shcolor"))                                            //  text shadow color
      zdialog_fetch(zd,"shcolor",attr.color[3],20);

   if (strEqu(event,"fgtransp"))                                           //  foreground transparency
      zdialog_fetch(zd,"fgtransp",attr.transp[0]);

   if (strEqu(event,"bgtransp"))                                           //  background transparency
      zdialog_fetch(zd,"bgtransp",attr.transp[1]);

   if (strEqu(event,"totransp"))                                           //  text outline transparency
      zdialog_fetch(zd,"totransp",attr.transp[2]);

   if (strEqu(event,"shtransp"))                                           //  text shadow transparency
      zdialog_fetch(zd,"shtransp",attr.transp[3]);

   if (strEqu(event,"towidth"))                                            //  text outline width
      zdialog_fetch(zd,"towidth",attr.towidth);

   if (strEqu(event,"shwidth"))                                            //  text shadow width
      zdialog_fetch(zd,"shwidth",attr.shwidth);

   if (strEqu(event,"shangle"))                                            //  text shadow angle
      zdialog_fetch(zd,"shangle",attr.shangle);

   attr.text = text;                                                       //  v.13.10
   gentext(&attr,pxm_text,pxm_transp);                                     //  build text image from text and attributes
   write(1);                                                               //  write on image
   zmainloop();

   return 1;
}


//  mouse function, set position for text on image

void writetext::mousefunc()
{
   using namespace writetext;

   if (LMclick) {                                                          //  left mouse click
      px = Mxclick;                                                        //  new text position on image
      py = Myclick;
      write(1);                                                            //  erase old, write new text on image
   }

   if (RMclick) {                                                          //  right mouse click
      write(2);                                                            //  erase old text, if any
      px = py = -1;                                                        //  v.12.03
   }

   if (Mxdrag || Mydrag)                                                   //  mouse dragged
   {
      px = Mxdrag;                                                         //  new text position on image
      py = Mydrag;
      write(1);                                                            //  erase old, write new text on image
   }

   LMclick = RMclick = Mxdrag = Mydrag = 0;                                //  v.12.11
   return;
}


//  write text on image at designated location
//  mode: 1  erase old and write to new position
//        2  erase old and write nothing

void writetext::write(int mode)                                            //  v.12.02
{
   using namespace writetext;

   float       *pix1, *pix2, *pix3;
   int         px1, py1, px3, py3, done;
   float       e3part;

   static int  orgx1, orgy1, ww1, hh1;                                     //  old text image overlap rectangle
   int         orgx2, orgy2, ww2, hh2;                                     //  new overlap rectangle

   if (present)                                                            //  v.13.03
   {
      for (py3 = orgy1; py3 < orgy1 + hh1; py3++)                          //  erase prior text image
      for (px3 = orgx1; px3 < orgx1 + ww1; px3++)                          //  replace E3 pixels with E1 pixels
      {                                                                    //    in prior overlap rectangle
         if (px3 < 0 || px3 >= E3pxm->ww) continue;
         if (py3 < 0 || py3 >= E3pxm->hh) continue;
         pix1 = PXMpix(E1pxm,px3,py3);
         pix3 = PXMpix(E3pxm,px3,py3);
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }
   }

   done = 0;
   if (mode == 2) done = 1;                                                //  erase only
   if (! *text) done = 1;                                                  //  no text defined
   if (px < 0 && py < 0) done = 1;                                         //  no position defined

   if (done) {
      if (present) {
         mwpaint3(orgx1,orgy1,ww1,hh1);                                    //  update window to erase old text
         present = 0;                                                      //  mark no text present
      }
      return;
   }

   ww2 = pxm_text->ww;                                                     //  text image buffer
   hh2 = pxm_text->hh;

   orgx2 = px - ww2/2;                                                     //  copy-to image3 location
   orgy2 = py - hh2/2;

   for (py1 = 0; py1 < hh2; py1++)                                         //  loop all pixels in text image
   for (px1 = 0; px1 < ww2; px1++)
   {
      px3 = orgx2 + px1;                                                   //  copy-to image3 pixel
      py3 = orgy2 + py1;

      if (px3 < 0 || px3 >= E3pxm->ww) continue;                           //  omit parts beyond edges
      if (py3 < 0 || py3 >= E3pxm->hh) continue;

      pix1 = PXMpix(pxm_text,px1,py1);                                     //  copy-from text pixel
      pix2 = PXMpix(pxm_transp,px1,py1);                                   //  copy-from transparency
      pix3 = PXMpix(E3pxm,px3,py3);                                        //  copy-to image pixel

      e3part = pix2[0] / 256.0;                                            //  image part visible through text

      pix3[0] = pix1[0] + e3part * pix3[0];                                //  combine text part + image part
      pix3[1] = pix1[1] + e3part * pix3[1];
      pix3[2] = pix1[2] + e3part * pix3[2];
   }

   if (present) {
      mwpaint3(orgx1,orgy1,ww1,hh1);                                       //  update window to erase old text
      present = 0;
   }
                                                                           //  (updates together to reduce flicker)
   mwpaint3(orgx2,orgy2,ww2,hh2);                                          //  update window for new text

   CEF->Fmods++;                                                           //  increase mod count           v.12.02
   CEF->Fsaved = 0;
   present = 1;                                                            //  mark text is present

   orgx1 = orgx2;                                                          //  remember overlap rectangle
   orgy1 = orgy2;                                                          //    for next call
   ww1 = ww2;
   hh1 = hh2;
   return;
}


/**************************************************************************/

//  Create a graphic image with text, white on black, any font, any angle.
//  Add outline and shadow colors. Apply transparencies.

int gentext(textattr_t *attr, PXM *&pxm_text, PXM *&pxm_transp)            //  v.13.10
{
   PXM * gentext_outline(textattr_t *, PXM *);
   PXM * gentext_shadow(textattr_t *, PXM *);

   PangoFontDescription    *pfont;
   PangoLayout             *playout;
   cairo_surface_t         *surface;
   cairo_t                 *cr;

   char           *text;
   float          angle;
   int            fgred, fggreen, fgblue;
   int            bgred, bggreen, bgblue;
   int            tored, togreen, toblue;
   int            shred, shgreen, shblue;
   float          fgtransp, bgtransp, totransp, shtransp;

   PXM            *pxm_temp1, *pxm_temp2, *pxm_temp3;
   uint8          *cairo_data, *cpix;
   float          *pix1, *pix2, *pix3;
   int            px, py, ww, hh;
   char           font[100];                                               //  font name and size
   int            red, green, blue;
   float          fgpart, topart, shpart, bgpart;

   zthreadcrash();                                                         //  thread usage not allowed

   text = attr->text;
   if (! text || ! *text) text = (char *) "null";                          //  v.13.10

   snprintf(font,100,"%s %d",attr->font,attr->size);                       //  font name and size together
   angle = attr->angle;                                                    //  text angle, degrees
   attr->sinT = sin(angle/57.296);                                         //  trig funcs for text angle
   attr->cosT = cos(angle/57.296);
   fgred = atoi(strField(attr->color[0],'|',1));                           //  get text foreground color
   fggreen = atoi(strField(attr->color[0],'|',2));
   fgblue = atoi(strField(attr->color[0],'|',3));
   bgred = atoi(strField(attr->color[1],'|',1));                           //  get text background color
   bggreen = atoi(strField(attr->color[1],'|',2));
   bgblue = atoi(strField(attr->color[1],'|',3));
   tored = atoi(strField(attr->color[2],'|',1));                           //  get text outline color
   togreen = atoi(strField(attr->color[2],'|',2));
   toblue = atoi(strField(attr->color[2],'|',3));
   shred = atoi(strField(attr->color[3],'|',1));                           //  get text shadow color
   shgreen = atoi(strField(attr->color[3],'|',2));
   shblue = atoi(strField(attr->color[3],'|',3));
   fgtransp = 0.01 * attr->transp[0];                                      //  get transparencies
   bgtransp = 0.01 * attr->transp[1];                                      //  text, background, outline, shadow
   totransp = 0.01 * attr->transp[2];
   shtransp = 0.01 * attr->transp[3];

   pfont = pango_font_description_from_string(font);                       //  make layout with text
   playout = gtk_widget_create_pango_layout(cDwin,null);
   pango_layout_set_font_description(playout,pfont);
   pango_layout_set_text(playout,text,-1);

   pango_layout_get_pixel_size(playout,&ww,&hh);
   ww += 2 + 0.2 * attr->size;                                             //  compensate bad font metrics
   hh += 2 + 0.1 * attr->size;
   attr->tww = ww;                                                         //  save text image size before rotate
   attr->thh = hh;

   surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,ww,hh);         //  cairo output image
   cr = cairo_create(surface);
   pango_cairo_show_layout(cr,playout);                                    //  write text layout to image
   cairo_data = cairo_image_surface_get_data(surface);                     //  output text image pixels

   pxm_temp1 = PXM_make(ww,hh);                                            //  create PXM

   for (py = 0; py < hh; py++)                                             //  copy text image to PXM
   for (px = 0; px < ww; px++)
   {
      cpix = cairo_data + 4 * (ww * py + px);                              //  pango output is monocolor
      pix2 = PXMpix(pxm_temp1,px,py);
      pix2[0] = cpix[3];                                                   //  use red [0] for text intensity
      pix2[1] = pix2[2] = 0;
   }

   pango_font_description_free(pfont);                                     //  free resources
   g_object_unref(playout);
   cairo_destroy(cr);
   cairo_surface_destroy(surface);

   pxm_temp2 = gentext_outline(attr,pxm_temp1);                            //  add text outline if any
   if (pxm_temp2) {                                                        //    using green [1] for outline intensity
      PXM_free(pxm_temp1);
      pxm_temp1 = pxm_temp2;
   }

   pxm_temp2 = gentext_shadow(attr,pxm_temp1);                             //  add text shadow color if any
   if (pxm_temp2) {                                                        //    using blue [2] for shadow intensity
      PXM_free(pxm_temp1);
      pxm_temp1 = pxm_temp2;
   }

   if (fabsf(angle) > 0.1) {                                               //  rotate text if wanted
      pxm_temp2 = PXM_rotate(pxm_temp1,angle);
      PXM_free(pxm_temp1);
      pxm_temp1 = pxm_temp2;
   }

   ww = pxm_temp1->ww;                                                     //  text image input PXM
   hh = pxm_temp1->hh;
   pxm_temp2 = PXM_make(ww,hh);                                            //  text image output PXM
   pxm_temp3 = PXM_make(ww,hh);                                            //  output transparency map

   for (py = 0; py < hh; py++)                                             //  loop all pixels in text image
   for (px = 0; px < ww; px++)
   {
      pix1 = PXMpix(pxm_temp1,px,py);                                      //  copy-from pixel (text + outline + shadow)
      pix2 = PXMpix(pxm_temp2,px,py);                                      //  copy-to pixel
      pix3 = PXMpix(pxm_temp3,px,py);                                      //  copy-to transparency

      fgpart = pix1[0] / 256.0;
      topart = pix1[1] / 256.0;
      shpart = pix1[2] / 256.0;
      bgpart = (1.0 - fgpart - topart - shpart);
      fgpart = fgpart * (1.0 - fgtransp);
      topart = topart * (1.0 - totransp);
      shpart = shpart * (1.0 - shtransp);
      bgpart = bgpart * (1.0 - bgtransp);

      red = fgpart * fgred + topart * tored + shpart * shred + bgpart * bgred;
      green = fgpart * fggreen + topart * togreen + shpart * shgreen + bgpart * bggreen;
      blue = fgpart * fgblue + topart * toblue + shpart * shblue + bgpart * bgblue;

      pix2[0] = red;                                                       //  output total red, green blue
      pix2[1] = green;
      pix2[2] = blue;

      pix3[0] = 255 * (1.0 - fgpart - topart - shpart - bgpart);           //  image part visible through text
   }

   PXM_free(pxm_temp1);
   pxm_text = pxm_temp2;
   pxm_transp = pxm_temp3;
   return 0;
}


//  add an outline color to the text character edges
//  red color [0] is original monocolor text
//  use green color [1] for added outline

PXM * gentext_outline(textattr_t *attr, PXM *pxm1)                         //  v.13.09
{
   PXM         *pxm2;
   int         toww, ww1, hh1, ww2, hh2;
   int         px, py, dx, dy;
   float       *pix1, *pix2, theta;

   toww = attr->towidth;                                                   //  text outline color width
   if (toww == 0) return 0;                                                //  zero

   ww1 = pxm1->ww;                                                         //  input PXM dimensions
   hh1 = pxm1->hh;
   ww2 = ww1 + toww * 2;                                                   //  add margins for outline width
   hh2 = hh1 + toww * 2;
   pxm2 = PXM_make(ww2,hh2);                                               //  output PXM

   for (py = 0; py < hh1; py++)                                            //  copy text pixels to outline pixels
   for (px = 0; px < ww1; px++)                                            //    displaced by outline width
   {
      pix1 = PXMpix(pxm1,px,py);
      pix2 = PXMpix(pxm2,px+toww,py+toww);
      pix2[0] = pix1[0];
   }

   theta = 0.7 / toww;
   for (theta = 0; theta < 6.3; theta += 0.7/toww)                         //  displace outline pixels in all directions
   {
      dx = roundf(toww * sinf(theta));
      dy = roundf(toww * cosf(theta));

      for (py = 0; py < hh1; py++)
      for (px = 0; px < ww1; px++)
      {
         pix1 = PXMpix(pxm1,px,py);
         pix2 = PXMpix(pxm2,px+toww+dx,py+toww+dy);
         if (pix2[1] < pix1[0] - pix2[0])                                  //  compare text to outline brightness
            pix2[1] = pix1[0] - pix2[0];                                   //  brighter part is outline pixel
      }
   }

   return pxm2;                                                            //  pix2[0] / pix2[1] = text / outline
}


//  add a shadow to the text character edges
//  red color [0] is original monocolor text intensity
//  green color [1] is added outline if any
//  use blue color [2] for added shadow

PXM * gentext_shadow(textattr_t *attr, PXM *pxm1)                          //  v.13.09
{
   PXM         *pxm2;
   int         shww, ww1, hh1, ww2, hh2;
   int         px, py, dx, dy;
   float       *pix1, *pix2, theta;

   shww = attr->shwidth;                                                   //  text shadow width
   if (shww == 0) return 0;                                                //  zero

   ww1 = pxm1->ww;                                                         //  input PXM dimensions
   hh1 = pxm1->hh;
   ww2 = ww1 + shww * 2;                                                   //  add margins for shadow width
   hh2 = hh1 + shww * 2;
   pxm2 = PXM_make(ww2,hh2);                                               //  output PXM

   for (py = 0; py < hh1; py++)                                            //  copy text pixels to shadow pixels
   for (px = 0; px < ww1; px++)                                            //    displaced by shadow width
   {
      pix1 = PXMpix(pxm1,px,py);
      pix2 = PXMpix(pxm2,px+shww,py+shww);
      pix2[0] = pix1[0];
      pix2[1] = pix1[1];
   }

   theta = (90 - attr->shangle) / 57.3;                                    //  degrees to radians, 0 = to the right
   dx = roundf(shww * sinf(theta));
   dy = roundf(shww * cosf(theta));

   for (py = 0; py < hh1; py++)                                            //  displace text by shadow width
   for (px = 0; px < ww1; px++)
   {
      pix1 = PXMpix(pxm1,px,py);
      pix2 = PXMpix(pxm2,px+shww+dx,py+shww+dy);
      if (pix2[2] < pix1[0] + pix1[1] - pix2[0] - pix2[1])                 //  compare text+outline to shadow pixels
         pix2[2] = pix1[0] + pix1[1] - pix2[0] - pix2[1];                  //  brighter part is shadow pixel
   }

   return pxm2;                                                            //  pix2[0] / pix2[1] / pix2[2]
}                                                                          //    = text / outline / shadow brightness


//  load text and attributes from a file

void load_text(zdialog *zd)                                                //  v.13.09
{
   FILE        *fid;
   int         err, nn;
   char        *pp, *pp2, *file, buff[1200];
   cchar       *dialogtitle = "load text data from a file";
   char        text[1000];                                                 //  text data
   textattr_t  attr;

   file = zgetfile(dialogtitle,"file",writetext_dirk);                     //  get input file from user
   if (! file) return;

   fid = fopen(file,"r");                                                  //  open for read
   if (! fid) {
      zmessageACK(mWin,0,"%s",strerror(errno));
      free(file);
      return;
   }

   pp = fgets_trim(buff,1200,fid);                                         //  read text string
   if (! pp) goto badfile;
   if (strnNeq(pp,"text string: ",13)) goto badfile;
   pp += 13;
   if (strlen(pp) < 2) goto badfile;
   repl_Nstrs(pp,text,"\\n","\n",null);                                    //  replace "\n" with newline char.

   pp = fgets_trim(buff,1200,fid);                                         //  read font and size
   if (! pp) goto badfile;
   if (strnNeq(pp,"font: ",6)) goto badfile;
   pp += 6;
   strTrim(pp);
   pp2 = strstr(pp,"size: ");
   if (! pp2) goto badfile;
   *pp2 = 0;
   pp2 += 6;
   strncpy0(attr.font,pp,80);
   attr.size = atoi(pp2);
   if (attr.size < 8) goto badfile;

   pp = fgets_trim(buff,1200,fid);                                         //  read text attributes
   if (! pp) goto badfile;
   nn = sscanf(pp,"attributes: %f  %s %s %s %s  %d %d %d %d  %d %d %d",
            &attr.angle, attr.color[0], attr.color[1], attr.color[2], attr.color[3],
            &attr.transp[0], &attr.transp[1], &attr.transp[2], &attr.transp[3],
            &attr.towidth, &attr.shwidth, &attr.shangle);
   if (nn != 12) goto badfile;

   err = fclose(fid);
   if (err) {
      zmessageACK(mWin,0,"%s",strerror(errno));
      free(file);
      return;
   }

   zdialog_stuff(zd,"text",text);                                          //  stuff zdialog fields
   zdialog_stuff(zd,"fontname",attr.font);
   zdialog_stuff(zd,"fontsize",attr.size);
   zdialog_stuff(zd,"angle",attr.angle);
   zdialog_stuff(zd,"fgcolor",attr.color[0]);
   zdialog_stuff(zd,"bgcolor",attr.color[1]);
   zdialog_stuff(zd,"tocolor",attr.color[2]);
   zdialog_stuff(zd,"shcolor",attr.color[3]);
   zdialog_stuff(zd,"fgtransp",attr.transp[0]);
   zdialog_stuff(zd,"bgtransp",attr.transp[1]);
   zdialog_stuff(zd,"totransp",attr.transp[2]);
   zdialog_stuff(zd,"shtransp",attr.transp[3]);
   zdialog_stuff(zd,"towidth",attr.towidth);
   zdialog_stuff(zd,"shwidth",attr.shwidth);
   zdialog_stuff(zd,"shangle",attr.shangle);
   return;

badfile:                                                                   //  project file had a problem
   fclose(fid);
   zmessageACK(mWin,0,ZTX("text file is defective"));
   printz("buff: %s\n",buff);
}


//  save text to a file

void save_text(zdialog *zd)                                                //  v.13.09
{
   cchar       *dialogtitle = "save text data to a file";
   FILE        *fid;
   char        *file, text[1000], text2[1200];
   textattr_t  attr;

   file = zgetfile(dialogtitle,"save",writetext_dirk);                     //  get output file from user
   if (! file) return;

   fid = fopen(file,"w");                                                  //  open for write
   if (! fid) {
      zmessageACK(mWin,0,"%s",strerror(errno));
      free(file);
      return;
   }

   zdialog_fetch(zd,"text",text,1000);                                     //  get text and attributes from zdialog
   zdialog_fetch(zd,"fontname",attr.font,80);
   zdialog_fetch(zd,"fontsize",attr.size);
   zdialog_fetch(zd,"angle",attr.angle);
   zdialog_fetch(zd,"fgcolor",attr.color[0],20);
   zdialog_fetch(zd,"bgcolor",attr.color[1],20);
   zdialog_fetch(zd,"tocolor",attr.color[2],20);
   zdialog_fetch(zd,"shcolor",attr.color[3],20);
   zdialog_fetch(zd,"fgtransp",attr.transp[0]);
   zdialog_fetch(zd,"bgtransp",attr.transp[1]);
   zdialog_fetch(zd,"totransp",attr.transp[2]);
   zdialog_fetch(zd,"shtransp",attr.transp[3]);
   zdialog_fetch(zd,"towidth",attr.towidth);
   zdialog_fetch(zd,"shwidth",attr.shwidth);
   zdialog_fetch(zd,"shangle",attr.shangle);

   repl_Nstrs(text, text2,"\n","\\n",null);                                //  replace newlines with "\n"

   fprintf(fid,"text string: %s\n",text2);

   fprintf(fid,"font: %s  size: %d \n",attr.font, attr.size);

   fprintf(fid,"attributes: %.4f  %s %s %s %s  %d %d %d %d  %d %d %d \n",
            attr.angle, attr.color[0], attr.color[1], attr.color[2], attr.color[3],
            attr.transp[0], attr.transp[1], attr.transp[2], attr.transp[3],
            attr.towidth, attr.shwidth, attr.shangle);

   fclose(fid);
   return;
}


/**************************************************************************
   plugin menu functions
**************************************************************************/

//  edit plugins menu

void  m_edit_plugins(GtkWidget *, cchar *)                                 //  simplified          v.13.12
{
   int   edit_plugins_event(zdialog *zd, cchar *event);

   int         ii;
   zdialog     *zd;
   
   F1_help_topic = "plugins";

/***
       _________________________________________
      |         Edit Plugins                    |
      |                                         |
      |  select [__________________________|v]  |
      |   edit  [____________________________]  |     Edit with Gimp = gimp
      |                                         |
      |                   [Add] [Remove] [Done] |
      |_________________________________________|

***/

   zd = zdialog_new("Edit Plugins",mWin,Badd,Bremove,Bdone,null);
   zdialog_add_widget(zd,"hbox","hbm","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labm","hbm",Bselect,"space=5");
   zdialog_add_widget(zd,"combo","select","hbm",0,"space=5");
   zdialog_add_widget(zd,"hbox","hbe","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labe","hbe",Bedit,"space=5");
   zdialog_add_widget(zd,"entry","edit","hbe",0,"space=5|expand");

   for (ii = 0; ii < Nplugins; ii++)                                       //  stuff combo box with available
      zdialog_cb_app(zd,"select",plugins[ii]);
   
   zdialog_run(zd,edit_plugins_event);
   return;
}


//  dialog event function

int  edit_plugins_event(zdialog *zd, cchar *event)
{
   int      ii, jj, cc, zstat;
   char     *pp, menuentry[200];
   char     buff[200];
   FILE     *fid;
   cchar    *errmess = ZTX("format: menu name = command %%s");

   if (strEqu(event,"select")) 
   {
      zdialog_fetch(zd,"select",menuentry,200);
      zdialog_stuff(zd,"edit",menuentry);                                  //  stuff corresp. command in dialog
      return 1;
   }
   
   zstat = zd->zstat;                                                      //  wait for dialog completion
   if (! zstat) return 0;
   
   if (zstat < 1 || zstat > 3) {                                           //  [x] button
      zdialog_free(zd);
      return 1;
   }
   
   if (zstat == 1)                                                         //  add new plugin
   {
      zd->zstat = 0;                                                       //  keep dialog active  bugfix      v.14.05

      if (Nplugins == maxplugins) {
         zmessageACK(mWin,0,"too many plugins");
         return 1;
      }
      
      zdialog_fetch(zd,"edit",menuentry,200);                              //  validate format:  menu = command %s
      pp = strstr(menuentry," = ");
      if (! pp || pp-menuentry < 3 || strlen(pp) < 6) {
         zmessageACK(mWin,0,errmess);
         return 1;
      }
      if (! strstr(pp+3," %s")) {
         zmessageACK(mWin,0,errmess);
         return 1;
      }
      
      cc = pp - menuentry + 1;                                             //  look for duplicate "menu name ="
      for (ii = 0; ii < Nplugins; ii++)
         if (strnEqu(plugins[ii],menuentry,cc)) break;
      if (ii < Nplugins) {
         zdialog_cb_delete(zd,"select",plugins[ii]);                       //  delete old entry from combo box
         Nplugins--;                                                       //  remove old plugin record        v.14.05
         for (jj = ii; jj < Nplugins; jj++)
            plugins[jj] = plugins[jj+1];
      }

      ii = Nplugins++;                                                     //  add plugin record
      plugins[ii] = strdup(menuentry);
      zdialog_cb_app(zd,"select",menuentry);                               //  add to combo box
      zdialog_stuff(zd,"select",menuentry);                                //  select this entry
   }
   
   if (zstat == 2)                                                         //  remove current plugin
   {
      zd->zstat = 0;                                                       //  keep dialog active              v.14.05

      zdialog_fetch(zd,"edit",menuentry,200);

      for (ii = 0; ii < Nplugins; ii++)
         if (strEqu(plugins[ii],menuentry)) break;
      if (ii == Nplugins) {
         zmessageACK(mWin,0,"menu not found");
         return 1;
      }
      
      Nplugins--;                                                          //  remove plugin record
      for (jj = ii; jj < Nplugins; jj++)
         plugins[jj] = plugins[jj+1];
      zdialog_cb_delete(zd,"select",menuentry);                            //  delete entry from combo box
   }
   
   if (zstat == 3)                                                         //  done
   {
      snprintf(buff,199,"%s/plugins",get_zuserdir());                      //  open file for plugins
      fid = fopen(buff,"w");
      if (! fid) {
         zmessageACK(mWin,0,strerror(errno));
         return 1;
      }
      
      for (int ii = 0; ii < Nplugins; ii++)                                //  save plugins
         if (plugins[ii])
            fprintf(fid,"%s \n",plugins[ii]);
      
      fclose(fid);

      zdialog_free(zd);
      zmessageACK(mWin,0,ZTX("Restart Fotoxx to update plugin menu"));
   }
   
   return 1;
}


//  process plugin menu selection
//  execute correspinding command using current image file

editfunc    EFplugin;

void  m_run_plugin(GtkWidget *, cchar *menu)
{
   int         ii, jj, err;
   char        *pp = 0, plugincommand[200], pluginfile[200];
   PXM         *pxmtemp;
   zdialog     *zd = 0;

   F1_help_topic = "plugins";

   if (checkpend("lock")) return;                                          //  check nothing pending              v.13.12

   for (ii = 0; ii < Nplugins; ii++)                                       //  search plugins for menu name
   {
      pp = strchr(plugins[ii],'=');                                        //  match menu name to plugin command
      if (! pp) continue;                                                  //  menu name = ...
      *pp = 0;
      jj = strEqu(plugins[ii],menu);
      *pp = '=';
      if (jj) break;
   }

   if (ii == Nplugins) {
      zmessageACK(mWin,0,"plugin menu not found %s",menu);
      return;
   }
   
   strncpy0(plugincommand,pp+1,200);                                       //  corresp. command
   strTrim2(plugincommand);

   pp = strstr(plugincommand,"%s");                                        //  bugfix                             v.14.05.1
   if (! pp) {
      zmessageACK(mWin,0,"plugin command has no \"%%s\" ");
      return;
   }

   EFplugin.menufunc = m_run_plugin;
   EFplugin.funcname = menu;
   if (! edit_setup(EFplugin)) return;                                     //  start edit function

   Fmenulock = 1;                                                          //  block others

   snprintf(pluginfile,199,"%s/plugfile.tif",tempdir);                     //  /.../fotoxx-nnnnnn/plugfile.tif
   
   err = PXM_TIFF_save(E1pxm,pluginfile,8);                                //  E1 >> plugin_file
   if (err) {
      if (*file_errmess)                                                   //  pass error to user
         zmessageACK(mWin,0,file_errmess);
      edit_cancel(0);
      goto RETURN;
   }

   zd = zmessage_post(mWin,0,ZTX("Plugin working ..."));                   //  v.13.12
   
   repl_1str(plugincommand,command,"%s",pluginfile);                       //  v.14.05

   err = shell_ack(command);                                               //  execute plugin command
   if (err) {
      edit_cancel(0);
      goto RETURN;
   }
   
   pxmtemp = TIFF_PXM_load(pluginfile);                                    //  read command output file
   if (! pxmtemp) {
      zmessageACK(mWin,0,ZTX("plugin failed"));
      edit_cancel(0);
      goto RETURN;
   }
   
   PXM_free(E3pxm);                                                        //  plugin_file >> E3
   E3pxm = pxmtemp;

   CEF->Fmods = 1;                                                         //  assume image was modified
   CEF->Fsaved = 0;
   edit_done(0);

RETURN:
   Fmenulock = 0;
   if (zd) zdialog_free(zd);
   return;
}   


